/*
 * Usually I write code in a 80x50 xterm (in Unix and Windows). 
 * But this module is written in a window of size 134x68, because
 * logging routine has long message etc and looks ungly in small window. 
 *
 * If you modify the code, please follow my coding style:
 *
 *   -- No Tabs please, use four spaces instead. 
 *   -- ANSI style bracing like I do.
 *   -- Declare variables like I do. 
 *
 * - muquit@muquit.com Jan-07-2004 
 */

/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

#include "apr_hooks.h"
#include "apr.h"
#include "apr_compat.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "apr_buckets.h"
#include "apr_md5.h"
#include "apr_network_io.h"
#include "apr_pools.h"
#include "apr_strings.h"
#include "apr_uri.h"
#include "apr_date.h"
#include "apr_fnmatch.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"

/*#define CORE_PRIVATE */
#include "ap_mpm.h" /* XXX */
#include "apr_thread_mutex.h"

#include <stdio.h>

#include <lber.h>
#include <ldap.h>

#ifdef HAVE_SSL
#include <ldap_ssl.h>
#endif /* HAVE_SSL */

#ifdef USE_MEM_CACHE
#include <ght_hash_table.h>
#endif /* USE_MEM_CACHE */

#ifdef USE_DATABASE
#include <ldbm.h>
#endif /* USE_DATABASE */

module AP_MODULE_DECLARE_DATA auth_ldap_module;


#define CACHE_DISK_DB_NAME      "mod_auth_ldap_cache.db"

#define MOD_AUTH_LDAP_VERSION_S "2.16"
#define MOD_AUTH_LDAP_SITE_URL  "http://muquit.com/"

/*
 * $Author: Muquit $
 * $Revision: 22 $
 * $Date: 5/23/04 10:51a $
 */

/*
**  mod_auth_ldap.c
**  Apache modue to use LDAP server for HTTP basic authentication.
**
**  Limitations and Comments:
**  It is simple and clean. works with Umich LDAP SDK and Netscape LDAP SDK
**
**  Development History:
**      who                  when           why
**   ma_muquit@fccc.edu      Sep-05-1998    first cut
**
**   ma_muquit@fccc.edu      Aug-19-1999    adding AuthLDAPAuthoritative
**        config file option, so that if the uid is not in the LDAP server,
**        control will be passed to lower modules.
**
**  Take.Vos@cable.a2000.nl  Sep-20-1999    fixed multiple groups free() too 
**       early. added partials DN to require group, instead of fixed group list
**
**  muquit@muquit.com        Feb-10-2000    replaced ap_getword() and my
**      rmallws() calls with  ap_getword_white(). Replaced ap_getword() and 
**      rmallws() calls for attributes with  ap_getword_conf() as 
**      suggested by Mark OLear <mgolear@ilstu.edu>. The reason to use
**      ap_getword_conf() was: if say require ou "Foo bar" is used, 
**      ap_geword() was tokenizing the string to Foo bar, but we want to
**      consider "Foo bar" as one string.
**
**  david@giffin.org         Jan-26-2001    Added a blind bind to see
**      if the user can bind with a dn of 'uid_attr=uid, base_dn' if
**      not we just keep going with the old scheme. This allows for
**      people to authenicate through a non-anonymous directory server
**      if it's set up consistantly. Also added a little better error when 
**      no user is found during the dn search.
**
**  muquit@muquit.com       Feb-03-2001     Added code to put LDAP_USER in the
**      environment if the user could authenticate successfully. Requested
**      by Ben Brewer <ben@polygon.net>.
**
**  muquit@muquit.com       Feb-05-2001     Possible memory leak. result in
**      ldap_search_s() was not freed by calling ldap_msgfree()
**
**  muquit@muquit.com       Mar-15-2001     Added Bind_DN, Bind_Pass. Added
**      env var MOD_AUTH_LDAP_VERSION. Release 2.3.
**
**  muquit@muquit.com       Mar-18-2001     added the filter option.
**
**  muquit@muquit.com       Apr-14-2001     The uids can have spaces now.
**      example: require user foo "john doe" bar 
**
**  muquit@muquit.com       Apr-11-2002     port to Apache 2.0 API
**
**  a.schild@aarboard.ch    Aug-09-2002     Implemented deref 
**       configuration options. Serialized ldap calls, since the ldap 
**       context handles are not reentrant. Added ldap_protocol_version stuff
**
**  muquit@muquit.com       Dec-22-2002     TLS support. Debug is a 
**                                          configurable option now.
**
**  muquit@muquit.com       Dec-24-2002     SSL support. a christmas eve hack
**      Coded according to sslnoauth.c comes with iplanet ldap C sdk example
**      code. Recquires iplanet C SDK and netscape navigator cert7.db with 
**      the CA certificate in it. 11:24 PM EST
**
** muquit@muquit.com        Feb-23-2003     Nested group support
** muquit@muquit.com        Mar-02-2003     new env var LDAP_USER_ALT
** I won't put comments here anymore.. May-26-2003 
**
**
* Tue Aug 26 2003 Scott Lampert <scott@lampert.org>
*
* - Added AuthOnBind apache config option
* - Added LDAP_Persistent apache config option
* - Added Bind_Tries apache config option
*
* $Id: mod_auth_ldap.c,v 1.3 2003/08/26 20:23:40 scott Exp $
*
*/

/* non ssl port */
#ifndef LDAP_PORT
#define LDAP_PORT   389
#endif /* LDAP_PORT */

#define MML    __LINE__
#define MFL    __FILE__,__LINE__

/*
 * Simple cache structure.
 * When a user successfully authenticates, we'll cache the following:
 *  - The ldap server used
 *  - The ldap port used
 *  - The IP address the user connected from. If using proxy, it'll always be the same. But that's ok.
 *  - MD5 crypted password. I could save the raw password, but if server cores for some reason, it's possible
 *    that the plain text password will show up in the core. So I'll use a md5 crypted version of the password in memory.
 *    If the password is good, it'll make dictionary attack harder. It'll be bit slower, but it's worth it I think.
 *  - The time the user is authenticated. It's the seconds from Unix epoch.
 *  When we lookup values from cache, the significant ones are, ldap server, ldap port, hashed password and last authentication time.
 *  If the user is not comping through a proxy, we'll re-authenticate if the TTL is expired. If the user is using proxy, the remote IP
 *  will be meaningless. But it's not that bad.
 */
typedef struct _CacheVal
{
    char
        ldap_server[128];   /* LDAP server user authenticated last */

    int
        ldap_port;          /* LDAP server port */

    char
        remote_ip[32];    /* remote IP in dotted quad */

    char
        cpw[128];          /* md5 crypted passwd */

    char
        user_dn[512];

    /* env var LDPA_USER_ALT */
    char
        ldap_user_alt[64];

    /* authorization */
    char
        filter[256];

    int
        filter_auth_passed;

    char
        group_dn[512];

    int
        group_auth_passed;

    char
        attr[32];

    int
        attr_auth_passed;
    /* authorization */

    time_t
        t;                 /* last time the user was authenticated */

    int
        pid;               /* the apache process PID */
}CacheVal;


typedef struct _LdapAuthSrvConf
{
    apr_thread_mutex_t
        *mutex;

    LDAP
        *ld;

    int
        ldap_persistent;

    int
        debug,
        use_cache,
        use_memory_cache,
        cache_ttl,
        use_disk_cache,
        hcount;

    int
        support_nested_groups;

#ifdef USE_MEM_CACHE
    ght_hash_table_t
        *hashp;
    int
        hash_table_size;
#endif /* USE_MEM_CACHE */


#ifdef USE_DATABASE
    char
        *cache_db_dir_path,
        *cache_db_path;
#endif /* USE_DATABASE */

}LdapAuthSrvConf;
static LdapAuthSrvConf *s_srv_conf;

#define TMPM_LOCK \
do \
{ \
    if (s_srv_conf->mutex) \
    { \
        apr_thread_mutex_lock(s_srv_conf->mutex); \
    } \
}while (0)

#define TMPM_UNLOCK \
do \
{ \
    if (s_srv_conf->mutex) \
    { \
        apr_thread_mutex_unlock(s_srv_conf->mutex); \
    } \
}while (0)


typedef struct _ldap_auth_config_rec
{
    LDAP
        *ld; /* The ldap connection. Must be protected by a semaphore*/

    char
        *ldap_server,
        *base_dn,
        *sub_dn,
        *uid_attr,
        *uid_attr_alt,
        *group_attr;

    char
        filter[512];

    char
        *user_dn;

    char
        *bind_dn,
        *bind_pass,
        *auth_on_bind;

    int
        ldap_port,
        ldap_protocol_version, /* The ldap protocoll version to use*/
        ldap_connect_timeout,
        use_ght_hashlib,
        max_bind_tries;

#ifdef USE_DATABASE
#endif /* USE_DATABASE */

    int
        env_vars_set,
        set_pass_to_env,
        debug,
        log_cache_stats,
        auth_ldapauthoritative;

#ifdef HAVE_LDAP_START_TLS_S
    int
        auth_ldapstarttls;
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
    char
        *ca_cert_db_path; /* path of cert7.db in format used by netscape communicator*/
#endif /* HAVE_SSL */
    int
        start_ssl;


    int
        ldap_opt_deref;

    char
        *dummy;

    int
        hcount;
} ldap_auth_config_rec;


/*
 * function prototypes. Strictly speaking I don't need them (because the way they're arranged).
 * But I want to see how many  functions I've in this module etc. I'll not declare apache module
 * related functions (like the ones to set directives etc).
 */
static void     m_rerr_va(request_rec *r,int debug,char *format,va_list args);
static void     m_rerr(request_rec *r,int debug,char *format,...);
static void     m_log_cache_stats(request_rec *r,int cache_stats,char *format,...);

LDAP            *mod_auth_ldap_initialize_ldap(request_rec *r);
static char     *mod_auth_ldap_ctime(time_t *t);
static void     mod_auth_ldap_free_dn(char *dn);
static char     *mod_auth_ldap_find_user_dn(request_rec *r,int *try_again);
static int      mod_auth_ldap_compare_group_recursively(request_rec *r,char *master_gr_dn);
static int      mod_auth_ldap_check_nested_groups(char *master_gr_dn,char *user_dn,request_rec *r);
static void     mod_auth_ldap_to_64(char *s, unsigned long v, int n);
static char     *mod_auth_ldap_set_ldap_user_alt(request_rec *r);

static int      mod_auth_ldap_insert_to_cache_from_authorization(CacheVal *cache_val,
                                                                 request_rec *r,char *group_dn,
                                                                 char *attr,char *filter);

static int      mod_auth_ldap_mod_auth_ldap_check_user_authorization_real(request_rec *r,char *group_dn,
                                                            char *attr,char *filter,
                                                            const char *value);

static int      mod_auth_ldap_find_user_in_memory_cache(request_rec *r,const char *sent_pw);

static int      mod_auth_ldap_insert_user_in_memory_cache(request_rec *r,const char *sent_pw,char *ldap_user_alt);

static int      mod_auth_ldap_find_user_in_disk_cache(request_rec *r,const char *sent_pw);

static int      mod_auth_ldap_insert_user_in_disk_cache(request_rec *r,const char *sent_pw,char *ldap_user_alt);

static int      mod_auth_ldap_delete_user_from_disk_cache(request_rec *r);

static int      mod_auth_ldap_delete_user_from_cache(request_rec *r);

static CacheVal *mod_auth_ldap_duplicate_cacheval(CacheVal *c,request_rec *r);
static CacheVal *mod_auth_ldap_get_cache_for_user(request_rec *r);



static void m_rerr_va(request_rec *r,int debug,char *format,va_list args)
{
    char
        buf[2048];
    int
        pid=getpid();

    if (debug)
    {
        ap_vsnprintf(buf,sizeof(buf),format,args);
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"(PID: %d) %s",pid,buf);
    }
}

/*
** Tired of cluttered log message with defines everywhere. A function call will
** be made even if DEBUG_LDAP is not defined. But I also don't have time
** to maintain two dlls for windows either. So debug is a configurable
** option now.
*/
static void m_rerr(request_rec *r,int debug,char *format,...)
{
    if (debug)
    {
        va_list
            args;

        char
            buf[2048];

        int
            pid=getpid();


        va_start(args,format);
        ap_vsnprintf(buf,2047,format,args); /* watch out..2047...avoiding calling sizeof all the time*/
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"(PID: %d) %s",pid,buf);
        va_end(args);
    }
}

static void m_serr(server_rec *s,int debug,char *format,...)
{

    if (debug)
    {
        va_list
            args;

        char
            buf[2048];

        int
            pid=getpid();

        va_start(args,format);
        ap_vsnprintf(buf,2047,format,args); /* watch out.. 2047.. avoiding calling sizeof all the time */
        ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,s,"(PID: %d) %s",pid,buf);
        va_end(args);
    }

}


static void m_log_cache_stats(request_rec *r,int log_cache_stats,char *format,...)
{
    va_list
        args;

    char
        buf[2048];

    int
        pid=getpid();

    if (log_cache_stats)
    {
        va_start(args,format);
        ap_vsnprintf(buf,2047,format,args); /* XXX */
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"(PID: %d) (CacheStats) %s",pid,buf);
        va_end(args);
    }
}



static void m_err(cmd_parms *cmd,int debug,char *format,...)
{
    va_list
        args;
    char
        buf[2048];

    if (debug)
    {
        va_start(args,format);
        ap_vsnprintf(buf,2047,format,args); /* XXX */
        ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,cmd->server,
                "%s",buf);
        va_end(args);
    }
}

/*
**  mod_auth_ldap_ctime()
**  get time, in the format, Wed Jun 30 21:49:08 1993
**
**  Parameters:
**  time_t  *t   - if not NULL use this time. if NULL use current time
**
**  Return Values:
**      pointer a string containing time
**
**  Comments:
**      remove the trailing new line from ctime(). subsequent call will
**      destroy the buffer. ctime() retuns pointer to a static buffer. So this routine is
**      is not thread safe. Here we're OK as we're mutexed when we call it.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    Jul-31-1999    first cut
*/
static char *mod_auth_ldap_ctime (time_t *t)
{
    time_t
        tm;

    char
        *times;

    if (t == NULL)
    {
        tm=time (NULL);
        times=ctime(&tm);
        times[(int) strlen(times)-1] = '\0';
    }
    else
    {
        times=ctime(t);
        times[(int) strlen(times)-1] = '\0';
    }
    return (times);
}


static void *create_ldap_auth_dir_config(apr_pool_t *p,char *d)
{
    ldap_auth_config_rec
        *cr;

    (void) srand((int) time((time_t *) NULL));

    cr=(ldap_auth_config_rec *) ap_pcalloc(p,sizeof(ldap_auth_config_rec));
    if (cr == (ldap_auth_config_rec *) NULL)
        return (NULL);

    memset(cr,0,sizeof(ldap_auth_config_rec));

    cr->ld=(LDAP *) NULL;

    cr->ldap_server=NULL; 
    cr->base_dn=NULL;
    cr->sub_dn=NULL;
    cr->user_dn=NULL;
    cr->bind_dn=NULL;
    cr->bind_pass=NULL;
    cr->max_bind_tries=10;
    cr->auth_on_bind=0;
    cr->uid_attr=ap_pstrdup(p,"uid");

    /*
     * Note about uid_attr_alt. If this attribute is specified, after
     * authenticating the use, get the value of that attribute and set it
     * to AUTH_USER env variable. Someone asked for it, because they use
     * that env var to match the ldap use to a user in some relational
     * database. Feb-23-2003 Feb-23-2003 
     */
    cr->uid_attr_alt=NULL;

    /*
     * Netscape/iPlanet LDAP server uses uniqueMember as the attribute to
     * store the user dn if the user is member of that group
     */
    cr->group_attr=ap_pstrdup(p,"uniqueMember");
    cr->ldap_port=LDAP_PORT;
    cr->ldap_protocol_version= LDAP_VERSION2;
    cr->ldap_connect_timeout=4; /* 4 seconds */

    cr->use_ght_hashlib=1; /* always */


    cr->env_vars_set=0;
    cr->debug=0;
    cr->log_cache_stats=0;
    cr->auth_ldapauthoritative=1; /* fortress is secure by default */
    cr->set_pass_to_env=0;

#ifdef HAVE_LDAP_START_TLS_S
    cr->auth_ldapstarttls=0;
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
    cr->ca_cert_db_path=NULL;
#endif /* HAVE_SSL */
    cr->start_ssl=0;
    cr->ldap_opt_deref= LDAP_DEREF_NEVER;  /* Default is to never deref */

    return (cr);
}

static const char *set_ldap_server(cmd_parms *cmd,void *dir_config,
                                   const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_server=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

#ifdef USE_DATABASE
static const char *set_cache_db_dir_path(cmd_parms *cmd,void *dir_config,
                                   const char *arg)
{
    s_srv_conf->cache_db_dir_path=ap_pstrdup(cmd->pool,arg);
    s_srv_conf->cache_db_path=apr_pstrcat(cmd->pool,s_srv_conf->cache_db_dir_path,"/",CACHE_DISK_DB_NAME,NULL);
    return (NULL);
}
#endif /* USE_DATABASE */

static const char *set_ldap_port(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_port=atoi(arg);
    return (NULL);
}

static const char *set_hash_table_size(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    /* only admin can set it.. so if she sets very large value, server will crush and burn! */
#ifdef USE_MEM_CACHE
    s_srv_conf->hash_table_size=atoi(arg);
#endif /* USE_MEM_CACHE */

    return (NULL);
}

static const char *set_cache_ttl(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    s_srv_conf->cache_ttl=atoi(arg);
    return (NULL);
}

static const char *set_ldap_protocol_version(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_protocol_version=atoi(arg);
    return (NULL);
}

static const char *set_ldap_connect_timeout(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_connect_timeout=atoi(arg);
    return (NULL);
}

static const char *set_ldap_deref(cmd_parms *cmd,void *dir_config,
                                  const char *arg)
{
    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *) dir_config;

    char
        *a=ap_pstrdup(cmd->pool,arg);

    if (a != NULL)
    {
        if ((strcasecmp(a,"NEVER") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is NEVER");
            cr->ldap_opt_deref= LDAP_DEREF_NEVER;
        }
        else if ((strcasecmp(a,"SEARCHING") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is SEARCHING");
            cr->ldap_opt_deref= LDAP_DEREF_SEARCHING;
        }
        else if ((strcasecmp(a,"FINDING") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is FINDING");
            cr->ldap_opt_deref= LDAP_DEREF_FINDING;
        }
        else if ((strcasecmp(a,"ALWAYS") == 0))
        {
           m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is ALWAYS");
           cr->ldap_opt_deref= LDAP_DEREF_ALWAYS;
        }
        else
        {
            m_err(cmd,cr->debug,
                    "[mod_auth_ldap.c] -  Deref has unexpected argument");
        }
    }
    else
    {
        m_err(cmd,cr->debug,
                "[mod_auth_ldap.c] - Deref is missing its argument");
    }
    return (NULL);
}

static const char *set_base_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->base_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_sub_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->sub_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_bind_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->bind_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_bind_pass(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->bind_pass=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_max_bind_tries(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->max_bind_tries=atoi(arg);

    if (cr->max_bind_tries <= 0) cr->max_bind_tries=1;
    return (NULL);
}

#ifdef HAVE_SSL
static const char *set_ca_cert_db_path(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *) dir_config;
    cr->ca_cert_db_path=ap_pstrdup(cmd->pool,arg);
    cr->start_ssl=1;
    return (NULL);
}
#endif /* HAVE_SSL */

static const char *set_uid_attr(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->uid_attr=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_uid_attr_alt(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->uid_attr_alt=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_group_attr(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->group_attr=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_ldap_persistent(cmd_parms *cmd,void *dummy,int arg)
{
    s_srv_conf->ldap_persistent=arg;
    return(NULL);

}

static const char *set_nested_groups(cmd_parms *cmd,void *dummy,int arg)
{
    s_srv_conf->support_nested_groups=arg;
    return(NULL);
}

static const char *set_use_memory_cache(cmd_parms *cmd,void *dummy,int arg)
{
    s_srv_conf->use_memory_cache=arg;
    return(NULL);
}

static const char *set_use_disk_cache(cmd_parms *cmd,void *dummy,int arg)
{
    s_srv_conf->use_disk_cache=arg;
    return(NULL);
}

static const char *set_ldap_debug(cmd_parms *cmd,void *dir_config,
                                  const char *arg)
{
    char
        *s;
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    if (arg)
    {
        if ((ap_strcasecmp_match(arg,"on") == 0) || (ap_strcasecmp_match(arg,"yes") == 0))
        {
            cr->debug=1;
            s_srv_conf->debug=1;
        }
    }

    return(NULL);
}

static const command_rec ldap_auth_cmds[]=
    {
        AP_INIT_TAKE1("LDAP_Debug",set_ldap_debug,
                     NULL,
                     OR_AUTHCFG, /* or should be restricted ?? */
                     "Print debug messages in server error log or not"),


        AP_INIT_FLAG("LDAP_LogCacheStats",ap_set_flag_slot,
                     (void *) APR_OFFSETOF(ldap_auth_config_rec,log_cache_stats),
                     RSRC_CONF,
                     "Print Cache statistics message to apache log. These messages can be parsed to show cache stats"),


        AP_INIT_TAKE1("LDAP_Server",set_ldap_server,
                      NULL,
                      OR_AUTHCFG,
                      "The name or IP address of the LDAP server"),

        AP_INIT_TAKE1("LDAP_Port",set_ldap_port,
                      NULL,
                      OR_AUTHCFG,
                      "The port ldap server is listening"),



        AP_INIT_TAKE1("LDAP_Protocol_Version",set_ldap_protocol_version,
                      NULL,
                      OR_AUTHCFG,
                      "The ldap protocol to use for communication with the ldap server"),

        AP_INIT_TAKE1("LDAP_Connect_Timeout",set_ldap_connect_timeout,
                      NULL,
                      OR_AUTHCFG,
                      "Timeout if could not connect in specified seconds"),


        AP_INIT_TAKE1("LDAP_Deref",set_ldap_deref,
                      NULL,
                      RSRC_CONF,
                      "The ldap dereferencing of alias"),

        AP_INIT_TAKE1("Base_DN",set_base_dn,
                      NULL,
                      OR_AUTHCFG,
                      "LDAP Base DN"),

        AP_INIT_TAKE1("Sub_DN",set_sub_dn,
                      NULL,
                      OR_AUTHCFG,
                      "Specify a node in the Base DN with this directive to construct a user DN for explicit binding if AuthOnBind is On"),

        AP_INIT_TAKE1("UID_Attr",set_uid_attr,
                      NULL,
                      OR_AUTHCFG,
                      "LAP attribute for uid"),


        AP_INIT_TAKE1("UID_Attr_Alt",set_uid_attr_alt,
                      NULL,
                      OR_AUTHCFG,
                      "LAP attribute to search and set to LDAP_USER_ALT after authentication"),


        AP_INIT_TAKE1("Group_Attr",set_group_attr,
                      NULL,
                      OR_AUTHCFG,
                      "Group attribute for user DN. iPlanet uses uniqueMember as the attribute"),

        AP_INIT_TAKE1("Bind_DN",set_bind_dn,
                      NULL,
                      OR_AUTHCFG,
                      "LDAP Bind DN"),

        AP_INIT_TAKE1("Bind_Pass",
                      set_bind_pass,NULL,
                      OR_AUTHCFG,
                      "LDAP Bind DN password"),

        AP_INIT_TAKE1("Bind_Tries",
	                  set_max_bind_tries,NULL,
                      OR_AUTHCFG,
		              "Maximum number of times to try to bind to the LDAP server"),

        AP_INIT_FLAG("AuthOnBind",ap_set_flag_slot,
                     (void *) APR_OFFSETOF(ldap_auth_config_rec,auth_on_bind),
                     OR_AUTHCFG,
                     "Set to On to consider a successful bind with the authenticating user's username + Base_DN and password to be a successful authentication."),

        AP_INIT_FLAG("LDAP_Persistent_G",set_ldap_persistent,NULL,
                     RSRC_CONF,
                     "Set to Off to make a new connection for every LDAP request."),

        AP_INIT_FLAG("SupportNestedGroups_G",set_nested_groups,NULL,
                     RSRC_CONF,
                     "Set to On if nested group support is needed. It's a expensive operation."),


        AP_INIT_FLAG("LDAP_UseMemoryCache_G",
                    set_use_memory_cache,NULL,
                    RSRC_CONF,
                    "Cache Authentication and Authorization info in memory or not. On or Off"),

        AP_INIT_TAKE1("LDAP_HashTableSize_G",set_hash_table_size,
                      NULL,
                      RSRC_CONF,
                      "Set the size of hash table to be used in memory cache"),


        AP_INIT_FLAG("LDAP_UseDiskCache_G",set_use_disk_cache,NULL,
                    RSRC_CONF,
                    "Cache Authentication and Authorization info in disk or not. On or Off"),

        AP_INIT_TAKE1("LDAP_CacheTTL_G",set_cache_ttl,
                      NULL,
                      RSRC_CONF,
                      "If LDAP_UseCache is on, specify cache TTL in seconds"),


#ifdef USE_DATABASE
        /* if LDAP_UseDiskCache_G is On, then path of cache db directory must be specified */
        AP_INIT_TAKE1("LDAP_CacheDbDirPath_G",set_cache_db_dir_path,
                      NULL,
                      RSRC_CONF,
                      "The full path of the directory of CACHE_DISK_DB_NAME, NOT the path of the database."),
#endif /* USE_DATABASE */

#ifdef HAVE_LDAP_START_TLS_S
        AP_INIT_FLAG("LDAP_StartTLS",ap_set_flag_slot,
                     (void *) APR_OFFSETOF(ldap_auth_config_rec,auth_ldapstarttls),
                     OR_AUTHCFG, 
                     "Set to yes to allow LDAP to enable Transport Layer Security"),
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
        AP_INIT_TAKE1("LDAP_CertDbDir",
                set_ca_cert_db_path,NULL,
                OR_AUTHCFG,
                "The full path of the directory containing certificates of Certificate Authorities in cert7.db and key3.db file used by Netscape Communicator"),
#endif /* HAVE_SSL */


        AP_INIT_FLAG("AuthLDAPAuthoritative",ap_set_flag_slot,
                     (void *) APR_OFFSETOF(ldap_auth_config_rec,auth_ldapauthoritative),
                     OR_AUTHCFG,
                     "Set to no to allow control to be passed along to lower modules if the uid s not known to this module"),

        AP_INIT_FLAG("LDAP_SetPasswordToEnv",ap_set_flag_slot,
                     (void *) APR_OFFSETOF(ldap_auth_config_rec,set_pass_to_env),
                     RSRC_CONF,
                     "Set to On if you need to set the password to become available in LDAP_PASSWORD env variable after the user is authenticated and authorized. The default is Off."),

        {NULL}
    };


static void mod_auth_ldap_free_dn(char *dn)
{
#if LDAP_SDK_VERSION == 2
        (void) free(dn);
#else
        ldap_memfree(dn);
#endif /* LDAP_SDK_VERSION != 2 */
}


#ifdef USE_MEM_CACHE
static apr_status_t mod_auth_ldap_cleanup_hash_table(void *conf)
{
    LdapAuthSrvConf
        *srv_conf;

    CacheVal
        *p_data;

    char
        *p_key;

    ght_iterator_t
        iterator;

    ght_hash_table_t
        *hashp;

    srv_conf=(LdapAuthSrvConf *) conf;
    if (srv_conf == NULL)
        return(APR_SUCCESS);

    hashp=srv_conf->hashp;
    if (hashp == NULL)
    {
        return(APR_SUCCESS);
    }

    /* free hash table */

    for (p_data = (CacheVal *)ght_first(hashp, &iterator, (void**)&p_key); p_data;
       p_data = (CacheVal *)ght_next(hashp, &iterator, (void**)&p_key))
    {
      /* Print out the entry */
      /* Free the data (the meta-data will be removed in ght_finalize below) */
#if 0
        (void) fprintf(stderr, "(insert)  ldap_server: %s\n",p_data->ldap_server);
        (void) fprintf(stderr, "(insert)    ldap_port: %d\n",p_data->ldap_port);
        (void) fprintf(stderr, "(insert)    remote IP: %s\n",p_data->remote_ip);
        (void) fprintf(stderr, "(insert)      user DN: %s\n",p_data->user_dn);
        (void) fprintf(stderr, "(insert)         pass: <not shown>\n");
        (void) fprintf(stderr, "(insert) Last Authenticated: %s\n",mod_auth_ldap_ctime(&p_data->t));
        (void) fflush(stderr);
#endif /* if 0 */

        (void)free((char *) p_data);

    }

    /* Remove the hash table */
    ght_finalize(hashp);


    return(APR_SUCCESS);
}
#endif /* USE_MEM_CACHE */


/**
 * configuation and start-up. 
 */
static int mod_auth_ldap_cache_post_config(apr_pool_t *p, apr_pool_t *plog,
        apr_pool_t *ptemp, server_rec *s)
{
    int
        threaded;

#ifdef USE_MEM_CACHE
    m_serr(s,s_srv_conf->debug,"%s (%d) - in mod_auth_ldap_cache_post_config(): %d,%d,%d",
            MFL,s_srv_conf->use_memory_cache,s_srv_conf->use_memory_cache,s_srv_conf->hash_table_size);

    ap_mpm_query(AP_MPMQ_IS_THREADED,&threaded);
    if (threaded)
    {
        apr_thread_mutex_create(&s_srv_conf->mutex,APR_THREAD_MUTEX_DEFAULT,p);
    }

    /*
     * create hash tbale only if: memory caching is ON, a hash table size 
     * is specified and if cache ttl is specified do not create hash table if disk caching is on
     */
    if (s_srv_conf->use_memory_cache && 
        s_srv_conf->hash_table_size  && 
        s_srv_conf->cache_ttl        && (!s_srv_conf->use_disk_cache))
    {
        m_serr(s,s_srv_conf->debug,"%s (%d) - Creating Hash table size: %d",MFL,s_srv_conf->hash_table_size);
        s_srv_conf->hashp=ght_create(s_srv_conf->hash_table_size);

        apr_pool_cleanup_register(p,s_srv_conf,mod_auth_ldap_cleanup_hash_table,apr_pool_cleanup_null);
        if (s_srv_conf->hashp)
        {
            m_serr(s,s_srv_conf->debug,"%s (%d) - hash table created: 0x%x",MFL,s_srv_conf->hashp);
            return(OK);
        }

        m_serr(s,s_srv_conf->debug,"%s (%d) - Could not create hash table of size %d",MFL,s_srv_conf->hash_table_size);
    }
    else
    {
        return(OK);
    }
#else
        return(OK);
#endif /* USE_MEM_CACHE */

    return(-1);
}

/*
**  mod_auth_ldap_find_user_dn()
**    return the DN of the user 
**
**  Parameters:
**      r           pointer to request_rec
**      try_again   OUT. if try_again is one, then the caller can retry to bind
**                       otherwise it means the connection is ok but the user may
**                       not exist.
**
**  Return Values:
**
**      pointer to DN of the userid if found, NULL otherwise
**
**  Limitations and Comments:
**      ld must be valid.
**
**      any LDAP sdk should work. returns pointer to a malloc'd space, the 
**      caller is responsible to free it.
**
**  Development History:
**      who                  when           why
**      ma_muquit@fccc.edu   Nov-29-1998    first cut for my own web server
**                                          mhttpd
**      muquit@muquit.com    Mar-15-2001    bind with bind and pass if provided.
*/
static char *mod_auth_ldap_find_user_dn(request_rec *r,int *try_again)
{
    int
    count = 0;

    int
        bound=0,
        rc=(-1);

    LDAPMessage
        *result,
        *entry;

    char
        *dn,
        szfilter[256];

    ldap_auth_config_rec
        *cr;

    char
        *base_dn,
        *uid_attrib,
        *userid,
        *bind_dn,
        *bind_pass;

    result=(LDAPMessage *) NULL;
    entry=(LDAPMessage *) NULL;
    dn=(char *) NULL;
    *try_again=1;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL || cr->ld == NULL)
        return(NULL);

    base_dn=cr->base_dn;
    uid_attrib=cr->uid_attr;
    userid=r->user;
    bind_dn=cr->bind_dn;
    bind_pass=cr->bind_pass;

    /* prepare filter with UidAttr. */
    ap_snprintf(szfilter,sizeof(szfilter)-1,"(%s=%s)",uid_attrib,userid);

    m_rerr(r,cr->debug,"%s (%d) - Using LDAP filter: %s",MFL,szfilter);

    /*
    ** if the bind DN and bind password is specified, try to bind with
    ** them first.
    */
    if (bind_dn != NULL && bind_pass != NULL)
    {
        m_rerr(r,cr->debug,"%s (%d) - Trying to bind with bind DN \"%s\" and password (not shown)",MFL,bind_dn);

        if (ldap_simple_bind_s(cr->ld,bind_dn,bind_pass) == LDAP_SUCCESS)
        {
            bound=1;
            m_rerr(r,cr->debug,"%s (%d) - Bound successfully with DN \"%s\" and password (not shown)",MFL,bind_dn);
        }
        else
        {
            m_rerr(r,1,"%s (%d) - Warning: bind with DN \"%s\" and password (not shown) did not succeed",MFL,bind_dn);
        }
    }

    /* Bind anonymously now only if not bound yet. */
    if (bound == 0)
    {
        /*
        rc=ldap_simple_bind_s(cr->ld,cr->base_dn,NULL);
        */
        m_rerr(r,1,"%s (%d) - cr->ld=%x",MFL,cr->ld);
        rc=ldap_simple_bind_s(cr->ld,NULL,NULL);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,
                "%s (%d) - Could not bind to LDAP server, reason: %s",
                    MFL,ldap_err2string(rc));
            return ((char *) NULL);
        }
        m_rerr(r,1,"%s (%d) - cr->ld=%x",MFL,cr->ld);
    }

    m_rerr(r,cr->debug,"%s (%d) - bound: %d, base dn: %s,filter: %s",
            MFL,bound,base_dn,szfilter);

    /* search to find the dn for the user */
    rc=ldap_search_s(cr->ld,
                     base_dn,
                     LDAP_SCOPE_SUBTREE,
                     szfilter,
                     NULL,
                     0,
                     &result);
    if (rc != LDAP_SUCCESS)
    {
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"%s (%d) - ldap_search_s() failed",MFL);
        /*
        ldap_perror(cr->ld,"ldap_search_s");
        (void) fflush(stderr);
        */

        /* note: ldap_err2string() returns pointer to a static space */
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                      "[mod_auth_ldap.c] - Error: %s",ldap_err2string(rc));
        return (NULL);
    }

    /* found something, better be a single match */
    if ((count = ldap_count_entries(cr->ld,result)) != 1)
    {
        if (count > 0)
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "[mod_auth_ldap.c] - multiple matches found for: %s",userid);
        }
        else
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "(PID: %d) %s (%d) - no user found for: %s",getpid(),MFL,userid);
        }

        /* don't try again, connection is open but user not found */
         *try_again=0;

        if (result != (LDAPMessage *) NULL)
        {
            ldap_msgfree(result);
            result=NULL;
        }

        return (NULL);
    }
    if ((entry =ldap_first_entry(cr->ld,result)) == NULL)
    {
        m_rerr(r,cr->debug,
         "[mod_auth_ldap.c] - %s=%s Unknown in LDAP server",uid_attrib,userid);

        if (result != (LDAPMessage *) NULL)
        {
            ldap_msgfree(result);
            result=NULL;
        }

        /* don't try again, connection is open but user not found */
        *try_again=0;
        return (NULL);
    }

    if ((dn=ldap_get_dn(cr->ld,entry)) == NULL)
    {
        m_rerr(r,1,"%s (%d) - could not obtain User DN from LDAP for user: %s",MFL,userid);
        return (NULL);
    }

    if (result != (LDAPMessage *) NULL) /* XXX */
    {
        ldap_msgfree(result);
        result=NULL;
    }

    return (dn); /* caller must free it */
}

static int mod_auth_ldap_compare_group_recursively(request_rec *r,char *master_gr_dn)
{
    char
        *gr_dn=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;

    int
        retval=(-1),
        rc;

    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    gr_dn=NULL;

    for(;;)
    {
        if (gr_dn)
        {
            mod_auth_ldap_free_dn(gr_dn);
            gr_dn=NULL;
        }

        if (res != NULL)
        {
            ldap_msgfree(res);
            res=NULL;
        }
        if (cr->debug)
        {
            m_rerr(r,cr->debug, "%s (%d) - mod_auth_ldap_compare_group_recursively(), filter=%s",MFL,cr->filter);
        }

        rc=ldap_search_s(cr->ld,
                cr->base_dn,
                LDAP_SCOPE_SUBTREE,
                cr->filter,
                NULL,
                0,
                &res);
        if (rc == LDAP_SUCCESS)
        {
            for (ent=ldap_first_entry(cr->ld,res); ent ;
                    ent=ldap_next_entry(cr->ld,ent))
            {
                gr_dn=ldap_get_dn(cr->ld,ent);
                if (gr_dn == NULL)
                {
                    retval=(-1);
                    goto exit_processing;
                }

                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "%s (%d) - >>Trying Group: %s",MFL,gr_dn);
                }

                rc=ldap_compare_s(cr->ld,
                        master_gr_dn,
                        cr->group_attr,
                        gr_dn);

                if (rc == LDAP_COMPARE_TRUE)
                {
                    m_rerr(r,cr->debug,
                        "%s (%d) - ** Group matched: %s",MFL,gr_dn);
                    mod_auth_ldap_free_dn(gr_dn);
                    gr_dn=NULL;

                    retval=(0);
                    goto exit_processing;
                }

                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "%s (%d) - No match",MFL);
                }

                ap_snprintf(cr->filter,sizeof(cr->filter)-1,
                    "(%s=%s)",cr->group_attr,gr_dn);
                res=NULL;
            }
            if (gr_dn == NULL)
            {
                retval=(-1);
                goto exit_processing;
            }
        }
        else
        {
            retval=(-1);
            goto exit_processing;
        }
    }

exit_processing:

    if (gr_dn)
    {
        mod_auth_ldap_free_dn(gr_dn);
        gr_dn=NULL;
    }

    if (res != NULL)
    {
        ldap_msgfree(res);
        res=NULL;
    }

    return(retval);
}



/**
 * return 0 if the user group belong to a master group
 *        -1 otherwise
 */
static int mod_auth_ldap_check_nested_groups(char *master_gr_dn,char *user_dn,request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        gr_count=0,
        count,
        retval=(-1),
        rc;

    char
        *dn=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    if (cr == NULL)
    {
        m_rerr(r,1,"%s (%d) - Could not get config rec\n",MFL);
        return(-1);
    }
    if (user_dn == NULL)
    {
        m_rerr(r,1,"%s (%d) - No user DN passed, will to check nested groups\n",MFL);
        return(-1);
    }

    /* find all the groups the user belongs to */
    ap_snprintf(cr->filter,sizeof(cr->filter)-1,"(%s=%s)",cr->group_attr,user_dn);
    m_rerr(r,cr->debug,"%s (%d) - using filter: %s",MFL,cr->filter);
    rc=ldap_search_s(cr->ld,
            cr->base_dn,
            LDAP_SCOPE_SUBTREE,
            cr->filter,
            NULL,
            0,
            &res);
    if (rc != LDAP_SUCCESS)
    {
        if (cr->debug)
        {
            m_rerr(r,cr->debug,
              "%s (%d) - mod_auth_ldap_check_nested_groups() user \"%s\" does not belong to any group",
                MFL,r->user);
        }
        retval=(-1);
        goto exit_processing;
    }
    count=ldap_count_entries(cr->ld,res);
    if (count <= 0)
    {
        if (cr->debug)
        {
            m_rerr(r,cr->debug,"%s (%d) - user \"%s\" does not belong to any group",MFL,r->user);
        }
        retval=(-1);
        goto exit_processing;
    }

    if (cr->debug)
    {
        m_rerr(r,cr->debug,"%s (%d) - user \"%s\" belongs to %d groups",
                MFL,r->user,count);
    }

    /*
    ** now go through each group and see if the group itself or if the group
    ** belong to any other group and that group belong some other group and 
    ** so on eventually belong to the master group
    */
    for (ent=ldap_first_entry(cr->ld,res); ent != NULL;
            ent=ldap_next_entry(cr->ld,ent))
    {
        gr_count++;
        dn=ldap_get_dn(cr->ld,ent);
        if (dn)
        {
            if (cr->debug)
            {
                m_rerr(r,cr->debug,"%s (%d) - >Trying (%d) Group: %s",MFL,gr_count,dn);
            }

            /* check if this group belong to the master group */
            rc=ldap_compare_s(cr->ld,
                    master_gr_dn,
                    cr->group_attr,
                    dn);

            if (rc == LDAP_COMPARE_TRUE)
            {
                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "%s (%d) - ** Group (%d) matched: %s",MFL,gr_count,dn);
                }
                mod_auth_ldap_free_dn(dn);
                dn=NULL;

                retval=(0);
                goto exit_processing;
            }
            else
            {
                if (cr->debug)
                {
                    m_rerr(r,cr->debug,"%s (%d) -  No match (%d)",MFL,gr_count);
                }
            }

            ap_snprintf(cr->filter,sizeof(cr->filter)-1,
                    "(%s=%s)",cr->group_attr,dn);

            mod_auth_ldap_free_dn(dn);
            dn=NULL;

            /*
            ** the sub group does not belong to the master group. check
            ** if the subgroup belongs to other groups which may belong to
            ** the master group
            */
            rc=mod_auth_ldap_compare_group_recursively(r,master_gr_dn);
            if (rc == 0)
            {
                retval=0;
                goto exit_processing;
            }
        }
    }

exit_processing:

    if (res != NULL)
        ldap_msgfree(res);

    return(retval);
}


static void mod_auth_ldap_to_64(char *s, unsigned long v, int n)
{
    static unsigned char itoa64[] =         /* 0 ... 63 => ASCII - 64 */
    "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

    while (--n >= 0) {
        *s++ = itoa64[v&0x3f];
        v >>= 6;
    }
}

/*
 * print the cache vals
 */
#ifdef USE_KAZLIB_HASH_X_X
static void mod_auth_ldap_dump_cache_val(request_rec *r)
{
    hscan_t
        hs;

    hnode_t
        *hn;

    CacheVal
        *cache_val;

    ldap_auth_config_rec
        *cr;

    const char
        *key;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);


    hash_scan_begin(&hs,s_srv_conf->hashp);
    while ((hn= hash_scan_next(&hs)))
    {
        key=hnode_getkey(hn);
        cache_val=hnode_get(hn);
        if (key && cache_val)
        {
            m_rerr(r,cr->debug,"(dump)         Hash dump,key: %s",key);
            m_rerr(r,cr->debug,"(dump)         Hash: %x, PID: %d",s_srv_conf->hashp,cache_val->pid);
            m_rerr(r,cr->debug,"(dump)  ldap_server: %s",cache_val->ldap_server);
            m_rerr(r,cr->debug,"(dump)    ldap_port: %d",cache_val->ldap_port);
            m_rerr(r,cr->debug,"(dump)    remote IP: %s",cache_val->remote_ip);
            /* m_rerr(r,cr->debug,"(dump)         pass: %s",cache_val->cpw); */
            m_rerr(r,cr->debug,"(dump)         pass: <not shown>");
            m_rerr(r,cr->debug,"(dump)Last authenticated: %s",
                    mod_auth_ldap_ctime(&cache_val->t));
        }
    }

}
#endif /* USE_KAZLIB_HASH_X */
/*
 * duplicate a CacheVal structre
 */
static CacheVal *mod_auth_ldap_duplicate_cacheval(CacheVal *c,request_rec *r)
{
    CacheVal
        *cache_val;;

    cache_val=(CacheVal *) malloc(sizeof(CacheVal));
    if (cache_val == (CacheVal *) NULL)
    {
        m_rerr(r,1,"%s (%d) - Could not duplicate cache, memrory allocation failed",MFL);
        return(NULL);
    }
    memset(cache_val,0,sizeof(CacheVal));

    /* populate our cache data */

    /* ldap server */
    if (*c->ldap_server != '\0')
        ap_snprintf(cache_val->ldap_server,sizeof(cache_val->ldap_server)-1,"%s",c->ldap_server);

    /* ldap port */
    cache_val->ldap_port=c->ldap_port;

    /* remote ip */
    if (*c->remote_ip != '\0')
        ap_snprintf(cache_val->remote_ip,sizeof(cache_val->remote_ip)-1,"%s",c->remote_ip);

    /* user DN */
    if (*c->user_dn != '\0')
        ap_snprintf(cache_val->user_dn,sizeof(cache_val->user_dn)-1,"%s",c->user_dn);

    /* md5 crypt the password */
    if (*c->cpw != '\0')
        ap_snprintf(cache_val->cpw,sizeof(cache_val->cpw)-1,"%s",c->cpw);

    /* time */
    cache_val->t=c->t;

    /* filter */
    if (*c->filter != '\0')
        ap_snprintf(cache_val->filter,sizeof(cache_val->filter)-1,"%s",c->filter);

    cache_val->filter_auth_passed=c->filter_auth_passed;

    /* group dn */
    if (*c->group_dn != '\0')
        ap_snprintf(cache_val->group_dn,sizeof(cache_val->group_dn)-1,"%s",c->group_dn);

    cache_val->group_auth_passed=c->group_auth_passed;

    /* LDAP attribute */
    if (*c->attr != '\0')
        ap_snprintf(cache_val->attr,sizeof(cache_val->attr)-1,"%s",c->attr);

    cache_val->attr_auth_passed=c->attr_auth_passed;

    cache_val->pid=c->pid;
    
    return(cache_val);
}

/**
 * lookup CacheVal for user in memory or in disk based db
 * returns pointer to CacheVal on success. Caller must free it.
 * NULL if Cache not found for the user
 */
static CacheVal *mod_auth_ldap_get_cache_for_user(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    CacheVal
        *cache_val=NULL;

#ifdef USE_DATABASE
    LDBM
        dbp=NULL;

    Datum
        key_data,
        return_data;

#endif /* USE_DATABASE */

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (s_srv_conf->use_memory_cache)
    {
#ifdef USE_MEM_CACHE
        m_rerr(r,cr->debug,"%s (%d) - Looking up cache for user: %s",
                MFL,r->user);
        cache_val=(CacheVal *) ght_get(s_srv_conf->hashp,strlen(r->user),r->user);
#endif /* USE_MEM_CACHE */
    }

    if (s_srv_conf->use_disk_cache)
    {
#ifdef USE_DATABASE
        dbp=ldbm_open(s_srv_conf->cache_db_path,LDBM_READER,00600,0);
        if (dbp == NULL)
        {
            m_rerr(r,1,"%s (%d) - Could not open database %s for reading",MFL,s_srv_conf->cache_db_path);
            return(NULL);
        }
        ldbm_datum_init(key_data);
        ldbm_datum_init(return_data);

        key_data.dptr=r->user;
        key_data.dsize=strlen(r->user);

        return_data=ldbm_fetch(dbp,key_data);
        if (return_data.dptr == NULL)
        {
            m_rerr(r,cr->debug,"%s (%d) - user %s is not found in disk cache",MFL,r->user);
            goto ExitProcessing;
        }
        cache_val=(CacheVal *) return_data.dptr; /* caller must free it */
#endif /* USE_DATABASE */
    }
    if (cache_val && cr->debug)
    {
        m_rerr(r,cr->debug,"%s (%d) - Found in Cache:",MFL);
#ifdef USE_MEM_CACHE
        m_rerr(r,cr->debug,"(found)         Hash: %x, PID: %d",s_srv_conf->hashp,cache_val->pid);
#endif /* USE_MEM_CACHE */
        m_rerr(r,cr->debug,"(found)  ldap_server: %s",cache_val->ldap_server);
        m_rerr(r,cr->debug,"(found)    ldap_port: %d",cache_val->ldap_port);
        m_rerr(r,cr->debug,"(found)    remote IP: %s",cache_val->remote_ip);
        m_rerr(r,cr->debug,"(found)      user DN: %s",cache_val->user_dn);
        m_rerr(r,cr->debug,"(found)         pass: <not shown>");
        if (cache_val->group_auth_passed)
        {
            m_rerr(r,cr->debug,"(found) Group DN: %s",cache_val->group_dn);
            m_rerr(r,cr->debug,"(found) group_auth_passed: %d",cache_val->group_auth_passed);
        }
        if (cache_val->attr_auth_passed)
        {
            m_rerr(r,cr->debug,"(found) LDAP Attr: %s",cache_val->attr);
            m_rerr(r,cr->debug,"(found) attr_auth_passed: %d",cache_val->attr_auth_passed);
        }
        if (cache_val->filter_auth_passed)
        {
            m_rerr(r,cr->debug,"(found) LDAP Filter: %s",cache_val->filter);
            m_rerr(r,cr->debug,"(found) filter_auth_passed: %d",cache_val->filter_auth_passed);
        }
        m_rerr(r,cr->debug,"(found) Last authenticated: %s",mod_auth_ldap_ctime(&cache_val->t));
    }

ExitProcessing:

#ifdef USE_DATABASE
    if (dbp)
        ldbm_close(dbp);
#endif /* USE_DATABASE */
    return(cache_val);
}

/**
 *  This routine is called after authorization by group or some kind of ldap
 *  attributes to insert the info to cache.
 *
 *  Params:
 *  cache_val   - Current cache
 *   r          - request rec
 *   group_dn   - Group DN if user is authorized by belonging to a group. NULL otherwise
 *   attr       - If user is authorized due to having a LDAP attribute. NULL otherwise
 *   filter     - if use is authorized by a filter. NULL otherwise
 *   Note: only of the above 3 will be not NULL.
 *  
 *  Gloabl changes: s_srv_conf->hashp
 *
 *  returns: 0 on success -1 on failure
 */
static int mod_auth_ldap_insert_to_cache_from_authorization(CacheVal *cache_val,request_rec *r,char *group_dn,char *attr,char *filter)
{
    int
        rc=(-1);

    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);


    if (cache_val == NULL)
        return(0);

#ifdef USE_MEM_CACHE
        if (s_srv_conf->use_memory_cache)
        {
                if (group_dn)
                {
                    ap_snprintf(cache_val->group_dn,sizeof(cache_val->group_dn)-1,"%s",group_dn);
                    cache_val->group_auth_passed=1;
                }

                if (attr)
                {
                    ap_snprintf(cache_val->attr,sizeof(cache_val->attr)-1,"%s",attr);
                    cache_val->attr_auth_passed=1;
                }

                if (filter)
                {
                    ap_snprintf(cache_val->filter,sizeof(cache_val->filter)-1,"%s",filter);
                    cache_val->filter_auth_passed=1;
                }
                if (group_dn)
                    m_rerr(r,cr->debug,"%s (%d) - Group dn: %s inserted to cache",MFL,group_dn);

                if (attr)
                    m_rerr(r,cr->debug,"%s (%d) - Attribute: %s inserted to cache",MFL,attr);

                if (filter)
                    m_rerr(r,cr->debug,"%s (%d) - Filter: %s inserted to cache",MFL,filter);


        }
#endif /* USE_MEM_CACHE */

#ifdef USE_DATABASE
        if (s_srv_conf->use_disk_cache)
        {
            LDBM
                dbp=NULL;

            Datum
                key_data,
                data_data;

            dbp=ldbm_open(s_srv_conf->cache_db_path,LDBM_WRCREAT,00600,0);
            if (dbp == NULL)
            {
                m_rerr(r,1,"%s (%d) - Could not open cache db: %s",MFL,s_srv_conf->cache_db_path);
                if (cache_val)
                    (void) free((char *) cache_val); /* XXX */
                rc=(-1);
                goto ExitProcessing;
            }
            ldbm_datum_init(key_data);
            ldbm_datum_init(data_data);

            if (group_dn)
            {
                ap_snprintf(cache_val->group_dn,sizeof(cache_val->group_dn)-1,"%s",group_dn);
                cache_val->group_auth_passed=1;
            }

            if (attr)
            {
                ap_snprintf(cache_val->attr,sizeof(cache_val->attr)-1,"%s",attr);
                cache_val->attr_auth_passed=1;
            }

            if (filter)
            {
                ap_snprintf(cache_val->filter,sizeof(cache_val->filter)-1,"%s",filter);
                cache_val->filter_auth_passed=1;
            }

            key_data.dptr=r->user;
            key_data.dsize=strlen(r->user);

            data_data.dptr=cache_val;
            data_data.dsize=sizeof(CacheVal);
            if (ldbm_store(dbp,key_data,data_data,LDBM_REPLACE) != 0)
            {
                m_rerr(r,1,"%s (%d) Could not save cache data for user to disk db",MFL);
                if (cache_val)
                    (void) free((char *) cache_val); /* XXX */
                ldbm_close(dbp);
                rc=(-1);
                goto ExitProcessing;
            }

            if (group_dn)
                m_rerr(r,cr->debug,"%s (%d) - Group dn: %s inserted to disk cache",MFL,group_dn);

            if (attr)
                m_rerr(r,cr->debug,"%s (%d) - Attribute: %s inserted to disk cache",MFL,attr);

            if (filter)
                m_rerr(r,cr->debug,"%s (%d) - Filter: %s inserted to disk cache",MFL,filter);


            if (cache_val)
                (void) free((char *) cache_val); /* XXX */
            ldbm_close(dbp);
            rc=(-1);
            goto ExitProcessing;
        }
#endif /* USE_DATABASE */

ExitProcessing:

    return(rc);

}


/**
 * The all doing routine for authorization. It checks authorization. If
 * caching is ON, after successfull authorization, update the cache.
 *
 *  Params:
 *  cache_val   - Current cache
 *   r          - request rec
 *   group_dn   - Group DN if user is authorized by belonging to a group. 
 *   NULL otherwise
 *
 *   attr       - If user is authorized due to having a LDAP attribute. 
 *   NULL otherwise
 *
 *   filter     - if use is authorized by a filter. NULL otherwise
 *
 *   Note: only one of the above 3 will be not NULL.
 *
 *   value      - for group authorization, it's the group name, for attribute 
 *                authorization is the value for the attribute
 *
 *  
 *  Gloabl changes: s_srv_conf->hashp
 *
 *  Note: it's a messy routine, it handles lots of complicated logic.
 *
 * returns, OK, HTTP_UNAUTHORIZED or DECLINED
 *
 */
static int mod_auth_ldap_mod_auth_ldap_check_user_authorization_real(request_rec *r,char *group_dn,char *attr,
        char *filter,const char *value)
{
    int
        from_cache=0,
        rcode=HTTP_UNAUTHORIZED,
        doing_what=0,
        insert_to_cache=1,
        rc=0;

    time_t
        user_dn_retrieved=0,
        os_now,
        time_diff;

    char
        *user_dn=NULL,
        *a=NULL;

    const char
        *v=NULL;

    CacheVal
        *cache_val=NULL;

    LDAP
        *ld=NULL;

    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *)
            ap_get_module_config(r->per_dir_config,&auth_ldap_module);

#define FROM_GROUP      0x01
#define FROM_ATTR       0x02
#define FROM_FILTER     0x03

    if (group_dn)
    {
        a=group_dn;
        doing_what=FROM_GROUP;
    }
    if (attr)
    {
        a=attr;
        v=value;
        doing_what=FROM_ATTR;
    }
    if (filter)
    {
        doing_what=FROM_FILTER;
        v=value;
    }

    /* check if caching is enabled. if so get the cache for the user */
    if (s_srv_conf->use_cache)
    {
        cache_val=mod_auth_ldap_get_cache_for_user(r);
        if (cache_val == NULL)
        {
            /*
             * we're in trouble. it should not be. becuase we're using cache, it's
             * most likely that the user is authenticated from cache, and we may not even
             * have any open ldap connections at this time (connection  
             * gets open during authenticatin).
             */
            m_rerr(r,cr->debug,"%s (%d) - ** No cache found for user %s, doing normal authorization",MFL,r->user);
            goto There;
        }

        /* good! cache found */
        os_now=time(NULL);
        time_diff=os_now - cache_val->t;

        if (time_diff > s_srv_conf->cache_ttl)
        {
             m_rerr(r,cr->debug,"%s (%d) - **** Cache expired (%d sec old), TTL is %d secs",MFL,time_diff,s_srv_conf->cache_ttl);
             m_rerr(r,cr->debug,"%s (%d) - Cache last inserted: %s",MFL,mod_auth_ldap_ctime(&cache_val->t));

             /* do regular authorization */
             goto There;
        }
        else
        {
            m_rerr(r,cr->debug,"%s (%d) - Authentication Cache not expired yet (%d sec old) TTL is %d secs",MFL,time_diff,s_srv_conf->cache_ttl);

            /* check if authorization information is in the cache or not. we know authentication infomation is */
            switch(doing_what)
            {
                case FROM_GROUP:
                {
                    if (cache_val->group_auth_passed)
                    {
                        m_rerr(r,cr->debug,"%s (%d) -       group DN: \"%s\"",MFL,group_dn);
                        m_rerr(r,cr->debug,"%s (%d) - cache group DN: \"%s\"",MFL,cache_val->group_dn);
                        if (ap_strcasecmp_match(group_dn,cache_val->group_dn) == 0)
                        {
                            m_rerr(r,cr->debug,"%s (%d) - Group DN matched with cached one, not connecting to LDAP to check",MFL);
                            from_cache=1;
                            insert_to_cache=0;
                            rcode=OK;
                            goto ExitProcessing;
                        }
                        else
                        {
                            m_rerr(r,cr->debug,"%s (%d) Authorization attribute did not match with cache",MFL);
                        }

                    }
                    else
                    {
                        m_rerr(r,cr->debug,"%s (%d) - Howerver user authorization info is not in cache, must contact LDAP!",MFL);
                    }
                    break;
                }

                case FROM_ATTR:
                {
                    if (cache_val->attr_auth_passed)
                    {
                        m_rerr(r,cr->debug,"%s (%d) -       attr: \"%s\"",MFL,attr);
                        m_rerr(r,cr->debug,"%s (%d) - cache attr: \"%s\"",MFL,cache_val->attr);
                        if (ap_strcasecmp_match(attr,cache_val->attr) == 0)
                        {
                            m_rerr(r,cr->debug,"%s (%d) - Attr matched with cached one, not connecting to LDAP to check",MFL);

                            from_cache=1;
                            insert_to_cache=0;
                            rcode=OK;
                            goto ExitProcessing;
                        }
                        else
                        {
                            m_rerr(r,cr->debug,"%s (%d) Authorization attribute did not match with cache",MFL);
                        }
                    }
                    else
                    {
                        m_rerr(r,cr->debug,"%s (%d) - Howerver user authorization info is not in cache, must contact LDAP!",MFL);
                    }

                    break;
                }

                case FROM_FILTER:
                {
                    if (cache_val->filter_auth_passed && cache_val->filter && filter)
                    {
                        m_rerr(r,cr->debug,"%s (%d) -       filter: \"%s\"",MFL,filter);
                        m_rerr(r,cr->debug,"%s (%d) - cache filter: \"%s\"",MFL,cache_val->filter);
                        if (ap_strcasecmp_match(filter,cache_val->filter) == 0)
                        {
                            m_rerr(r,cr->debug,"%s (%d) - Filter matched with cached one, not connecting to LDAP to check",MFL);

                            from_cache=1;
                            insert_to_cache=0;

                            rcode=OK;
                            goto ExitProcessing;
                        }
                        else
                        {
                            m_rerr(r,cr->debug,"%s (%d) - Howerver user authorization info is not in cache, must contact LDAP!",MFL);
                        }

                    }
                    break;
                }

                default:
                {
                    m_rerr(r,cr->debug,"%s (%d) - unknown aothorization type",MFL);
                    rcode=HTTP_UNAUTHORIZED;
                    goto ExitProcessing;
                    break;
                }
            }
        }
    }

There:

    cr->user_dn=NULL; /* we'll not use from cr anymore */
    if (! s_srv_conf->ldap_persistent)
    {

        m_rerr(r,cr->debug,"%s (%d) - ** Initializing LDAP **",MFL);
        ld=mod_auth_ldap_initialize_ldap(r);
        /*ld=ldap_init(cr->ldap_server,cr->ldap_port);*/
        if (ld == NULL)
        {
            m_rerr(r,cr->debug,"%s (%d) - Could not initialize LDAP",MFL);
            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }
        m_rerr(r,cr->debug,"%s (%d) - ** Initialized LDAP **",MFL);
        cr->ld=ld;
    }
    /*
     * Looks like sometimes cr->user_dn gets messed up. There must be a race condition in
     * apache. Probably it is getting freed from pool too early?
     */
    if (cr->user_dn) /* never be */
    {
        user_dn=cr->user_dn;
        m_rerr(r,cr->debug,"%s (%d) - User dn: \"%s\"",MFL,user_dn);
        m_rerr(r,cr->debug,"%s (%d) - from_cache=%d",MFL,from_cache);
        /*
         * check if it is same as the one in the cache. It always should match. but if it 
         * doesn't, we'll use the one in the cache because it is always correct!
         */
        if (s_srv_conf->use_cache && cache_val && (*cache_val->user_dn != '\0'))
        {
            if (ap_strcasecmp_match(cache_val->user_dn,cr->user_dn) != 0)
            {
                m_rerr(r,cr->debug,"%s (%d) - *** user dn different in cr and cache, but it can't be!!",MFL);
                m_rerr(r,cr->debug,"%s (%d) - in cache: %s",MFL,cache_val->user_dn);
                m_rerr(r,cr->debug,"%s (%d) -   in  cr: %s",MFL,cr->user_dn);
                user_dn=cache_val->user_dn;
            }
        }
        else
        {
            if (!from_cache)
            {
                int
                    try_again=0;
                /* channce is user dn is screwed up */
                m_rerr(r,cr->debug,"%s (%d) - ** from_cache=0,retrieving user DN from LDAP **",MFL);
                if (cr->ld == NULL)
                {
                    m_rerr(r,1,"%s (%d) - No open LDAP connection during authorization",MFL);
                    rcode=HTTP_UNAUTHORIZED;
                    goto ExitProcessing;
                }

                user_dn=mod_auth_ldap_find_user_dn(r,&try_again);
                if (user_dn == NULL)
                {
                    m_rerr(r,cr->debug,"%s (%d) - Could not retrienve user DN for user %s from LDAP",MFL,r->user);
                    rcode=HTTP_UNAUTHORIZED;
                    goto ExitProcessing;
                }
                user_dn_retrieved=1; /* indicate we must free it */
                m_rerr(r,cr->debug,"%s (%d) - retrieved user dn: %s",MFL,user_dn);

            }
        }
    }
    else
    {
        int
            try_again=0;
        if (cr->ld == NULL)
        {
            m_rerr(r,1,"%s (%d) - No open LDAP connection during authorization",MFL);
            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }
        if (from_cache)
        {
            user_dn=cache_val->user_dn;
            m_rerr(r,cr->debug,"%s (%d) - using user dn from cache: %s",MFL,user_dn);
        }
        else
        {
            m_rerr(r,cr->debug,"%s (%d) - ** Retrieving user DN from LDAP **",MFL);
            user_dn=mod_auth_ldap_find_user_dn(r,&try_again);
            if (user_dn == NULL)
            {
                 m_rerr(r,cr->debug,"%s (%d) - Could not retrienve user DN for user %s from LDAP",MFL,r->user);
                 rcode=HTTP_UNAUTHORIZED;
                 goto ExitProcessing;
            }
            user_dn_retrieved=1; /* indicate we must free it */
            m_rerr(r,cr->debug,"%s (%d) - retrieved user dn: %s",MFL,user_dn);
        }
    }

    /* now we surely have a User DN */
    switch(doing_what)
    {
        case FROM_GROUP:
        {
            m_rerr(r,cr->debug,"%s (%d) - Checking (from LDAP) if user %s belongs to group %s",MFL,r->user,group_dn);
            if (cr->ld == NULL)
            {
                m_rerr(r,1,"%s (%d) - No open LDAP connection during authorization",MFL);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            m_rerr(r,cr->debug,"%s (%d) - Checking Group: \"%s\",group attr: \"%s\", user dn: \"%s\"",
                    MFL,group_dn,cr->group_attr,user_dn);

            rc=ldap_compare_s(cr->ld,
                    group_dn,
                    cr->group_attr,
                    user_dn); /* someone reported for openldap schema the value is: cr->user */
            if (rc == LDAP_COMPARE_TRUE)
            {
                m_rerr(r,cr->debug,"%s (%d) - user %s belongs to group \"%s\" directly",MFL,r->user,value);
                rcode=OK;
                goto ExitProcessing;
            }
            else if (rc == LDAP_COMPARE_FALSE)
            {
                m_rerr(r,cr->debug,"%s (%d) - ld: %x, user %s does not belong to group %s, (error: %s)",
                        MFL,cr->ld,r->user,value,ldap_err2string(rc));
                if (s_srv_conf->support_nested_groups)
                {
                    m_rerr(r,cr->debug,"%s (%d) - Checking nested groups",MFL);
                    rc=mod_auth_ldap_check_nested_groups(group_dn,user_dn,r);
                    if (rc == 0)
                    {
                        m_rerr(r,cr->debug,"%s (%d) - user %s belongs to group \"%s\" by nested groups",MFL,r->user,value);
                        rcode=OK;
                        goto ExitProcessing;
                    }
                }
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            else
            {
                m_rerr(r,cr->debug,"%s (%d) - Error searching for group (error: %s)",MFL,ldap_err2string(rc));
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }

            break;
        }

        case FROM_ATTR:
        {
            m_rerr(r,cr->debug,"%s (%d) - Checking (from LDAP) if user %s has attribute %s",MFL,r->user,attr);
            if (cr->ld == NULL)
            {
                m_rerr(r,1,"%s (%d) - No open LDAP connection during authorization",MFL);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }

            rc=ldap_compare_s(cr->ld,
                    user_dn,
                    attr,
                    value);
            if (rc == LDAP_COMPARE_TRUE)
            {
                m_rerr(r,cr->debug,"%s (%d) - ** OK user \"%s\" has attribute %s=%s",
                        MFL,r->user,attr,value);
                rcode=OK;
                goto ExitProcessing;
            }
            else if (rc == LDAP_COMPARE_FALSE)
            {
                m_rerr(r,cr->debug,"%s (%d) - **XX** user %s does not have attribute %s=%s",MFL,r->user,attr,value);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            else
            {
                m_rerr(r,cr->debug,"%s (%d) - LDAP error (%s)",MFL,ldap_err2string(rc));
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }

            break;
        }

        case FROM_FILTER:
        {
            LDAPMessage
                *result;

            m_rerr(r,cr->debug,"%s (%d) - Checking (from LDAP) if user %s passes the ldap filter %s",MFL,r->user,filter);
            if (cr->ld == NULL)
            {
                m_rerr(r,1,"%s (%d) - No open LDAP connection during authorization",MFL);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }

            rc=ldap_search_s(cr->ld,
                    user_dn,
                    LDAP_SCOPE_SUBTREE,
                    filter,
                    NULL,
                    0,
                    &result);
            if (rc != LDAP_SUCCESS)
            {
                if (result) /* XXX */
                    ldap_msgfree(result);

                m_rerr(r,cr->debug,"%s (%d) - ** Filter \"%s\" did not pass for user %s\n",MFL,filter,r->user);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            if (ldap_first_entry(cr->ld,result) == NULL)
            {
                if (result)
                    ldap_msgfree(result);
                m_rerr(r,cr->debug,"%s (%d) - ** Filter \"%s\" did not pass for user %s\n",MFL,filter,r->user);
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            if (rc == LDAP_SUCCESS && result)
            {
                m_rerr(r,cr->debug,"%s (%d) - OK Authorization passed, filter %s passed for user %s",MFL,filter,r->user);
                ldap_msgfree(result);
                rcode=OK;
                goto ExitProcessing;
            }
            if (result)
                ldap_msgfree(result);

            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;

            break;
        }
    }

   

ExitProcessing:

    if (user_dn_retrieved && user_dn)
    {
        mod_auth_ldap_free_dn(user_dn);
    }

    if (rcode == OK)
    {
        /* insert the authorization info to cache */
        m_rerr(r,cr->debug,"%s (%d) - insert_to_cache=%d,s_srv_conf->use_cache=%d,cache_val=0x%04x",
                MFL,insert_to_cache,s_srv_conf->use_cache,cache_val);

        if (insert_to_cache && s_srv_conf->use_cache && cache_val)
        {
            m_rerr(r,cr->debug,"%s (%d) - Inserting user authorization info to cache",MFL);
            (void) mod_auth_ldap_insert_to_cache_from_authorization(cache_val,r,group_dn,attr,filter);
        }
    }
    else
    {
        /*
         * don't delete cache, because the user is already authenticated, but authorization
         * has failed. If multiple group, attribute or filter is in the httpd.conf file,
         * one of them might succeed. If we delete the user will not be authorized from
         * cache if she is in the last group. e.g.
         * require roomnumber "foo"
         * require roomnumber "bar"
         * if user has attribtue bar then cache will get cleared in foo
         */
#if 0
        if (cache_val)
        {
            /*
             * user is authenticated and info is inserted to cache. But authorization
             * failed, so get rid of the cached info.
             */
             m_rerr(r,cr->debug,"%s (%d) - Deleting user from cache",MFL);
             mod_auth_ldap_delete_user_from_cache(r);
        }
#endif /* if 0 */

        if (! cr->auth_ldapauthoritative)
        {
            /* give control to lower level modules */
            rcode=DECLINED;
        }
    }
 

    if (rcode == OK)
    {
        m_rerr(r,cr->debug,"%s (%d) - user %s is Authorized (from %s)",MFL,r->user,from_cache ? "Cache" : "LDAP");
        if (cr->log_cache_stats && s_srv_conf->use_cache && s_srv_conf->use_memory_cache)
        {
#ifdef USE_MEM_CACHE
            m_log_cache_stats(r,cr->log_cache_stats,
                        "%s (%d) -  Authorization Hash table size:%d No of entries: %d",MFL,ght_tablesize(s_srv_conf->hashp),ght_size(s_srv_conf->hashp));
#endif /* USE_MEM_CACHE */
        }
    }

    if (! s_srv_conf->ldap_persistent)
    {
        if (ld && cr->ld)
        {
            m_rerr(r,cr->debug,"%s (%d) - ** Calling ldap_unbind %x,%x **",MFL,ld,cr->ld);
            ldap_unbind(cr->ld);
            m_rerr(r,cr->debug,"%s (%d) - ** Done Calling ldap_unbind **",MFL);
            cr->ld=NULL;
        }
    }
    return(rcode);
}


#if defined(USE_MEM_CACHE) 
/*
 * Check cache if the user data exists. If user data is in the cache but it has expired, 
 * delete the hash node and free cached data.
 * 
 * return 1 if a user is found in cahce, 0 otherwise
 */
static int mod_auth_ldap_find_user_in_memory_cache(request_rec *r,const char *sent_pw)
{
    char
        *user;

    ldap_auth_config_rec
        *cr;

    CacheVal
        *cache_val;

    /*
    apr_time_t
        timediff,
        now;
    */
    time_t
        timediff,
        os_now;

    int
        rc=1;

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL)
        return(0);

    if (r->user == NULL)
        return(0);
    user=r->user;

    if (cr->use_ght_hashlib)
    {
        cache_val=(CacheVal *) ght_get(s_srv_conf->hashp,strlen(user),user);
        if (cache_val)
        {
            /* user found in cache! validate data */
            if (cr->debug)
            {
                m_rerr(r,cr->debug,"%s (%d) - Found in Cache:",MFL);
                m_rerr(r,cr->debug,"(found)         Hash: %x, PID: %d",s_srv_conf->hashp,cache_val->pid);
                m_rerr(r,cr->debug,"(found)  ldap_server: %s",cache_val->ldap_server);
                m_rerr(r,cr->debug,"(found)    ldap_port: %d",cache_val->ldap_port);
                m_rerr(r,cr->debug,"(found)    remote IP: %s",cache_val->remote_ip);
                m_rerr(r,cr->debug,"(found)      user DN: %s",cache_val->user_dn);
                /*m_rerr(r,cr->debug,"(found)         pass: %s",cache_val->cpw);*/
                m_rerr(r,cr->debug,"(found)         pass: <not shown>");
                m_rerr(r,cr->debug,"(found)  Last Authenticated: %s",mod_auth_ldap_ctime(&cache_val->t));
            }

            /* Verify with data from connection */

            if (ap_strcasecmp_match(cr->ldap_server,cache_val->ldap_server) != 0)
            {
                /* ldap server not the same */
                m_rerr(r,cr->debug,"%s (%d) - Cache: ldap servers not the same, conn: %s, cache: %s",MFL,cr->ldap_server,cache_val->ldap_server);
                rc=0;
                goto ExitProcessing;
            }
            m_rerr(r,cr->debug,"%s (%d) - ldap server matched with cached one, conn: %s, cahce: %s",MFL,cr->ldap_server,cache_val->ldap_server);

            if (cr->ldap_port != cache_val->ldap_port)
            {
                m_rerr(r,cr->debug,"%s (%d) - Cache: ldap port not the same, conn: %d, cache: %d",MFL,cr->ldap_port,cache_val->ldap_port);
                /* ldap port not the same */
                rc=0;
                goto ExitProcessing;
            }
            m_rerr(r,cr->debug,"%s (%d) - ldap port matched with cached one, conn: %d, cache: %d",MFL,cr->ldap_port,cache_val->ldap_port);

            if (ap_strcasecmp_match(r->connection->remote_ip,cache_val->remote_ip) != 0)
            {
                /* connected from different machine */
                m_rerr(r,cr->debug,"%s (%d) - Cache: remote ip did not match, conn: %s, cahce: %s",
                        MFL,r->connection->remote_ip,cache_val->remote_ip);
                rc=0;
                goto ExitProcessing;
            }

            m_rerr(r,cr->debug,"%s (%d) - remote IP matched with cached one, conn: %s, cahce: %s",
                    MFL,r->connection->remote_ip,cache_val->remote_ip);


            if (apr_password_validate(sent_pw,cache_val->cpw) != 0)
            {
                m_rerr(r,cr->debug,"%s (%d) - Cache: passwords do no match",MFL);
                /* password does not match */
                rc=0;
                goto ExitProcessing;
            }
            m_rerr(r,cr->debug,"%s (%d) - password matched with cached one",MFL);

            /*
            now=apr_time_now();
            */
            /*
            time(&os_now);
            timediff=os_now - (cache_val->t / APR_USEC_PER_SEC);
            */
            os_now=time(NULL);
            timediff=os_now - cache_val->t;
            if (timediff > s_srv_conf->cache_ttl)
            {
                m_rerr(r,cr->debug,"%s (%d) - **** Cache expired (%d sec old), TTL=%d",MFL,timediff,s_srv_conf->cache_ttl);
                m_rerr(r,cr->debug,"%s (%d) - Cache last inserted: %s",MFL,mod_auth_ldap_ctime(&cache_val->t));
                /* autheticated too long ago */
                rc=0;
                goto ExitProcessing;
            }
            m_rerr(r,cr->debug,"%s (%d) - Cache not expired yet (%d sec old), TTL=%d",MFL,timediff,s_srv_conf->cache_ttl);
            rc=1;

            /* set env vars */
            ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",MOD_AUTH_LDAP_VERSION_S);
            ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_CACHE_TYPE","Memory");
            ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
            if (*cache_val->ldap_user_alt)
            {
                ap_table_setn(r->subprocess_env,"LDAP_USER_ALT",cache_val->ldap_user_alt);
            }

            /* not secure but do it if you want, some people asked for it */
            if (cr->set_pass_to_env && sent_pw)
            {
                ap_table_setn(r->subprocess_env,"LDAP_PASSWORD",sent_pw);
            }
        }
        else
        {
            m_rerr(r,cr->debug,"%s (%d) - User \"%s\" is not found in cache",MFL,user);
            m_rerr(r,cr->debug,"         Hash: %x",s_srv_conf->hashp);
            return(0);
        }
    }

ExitProcessing:

    if (rc == 0)
    {
        if (cr->use_ght_hashlib)
        {
            cache_val=ght_remove(s_srv_conf->hashp,strlen(user),user);
            if (cache_val)
            {
                m_rerr(r,cr->debug,"%s (%d) - Deleted user %s from cache",MFL,user);
                (void) free((char *) cache_val);
            }
            else
            {
                m_rerr(r,cr->debug,"%s (%d) - Key not found for user %s to delete from cahce",MFL,user);
            }
        }
    }

    return(rc);
}

/*
 * Insert user data to cache. This routine is called after the user is authenticated
 * returns 0 on sucess -1 otherwise
 */
static int mod_auth_ldap_insert_user_in_memory_cache(request_rec *r,const char *sent_pw,char *ldap_user_alt)
{
    char
        ubuf[64],
        *user=NULL;

    ldap_auth_config_rec
        *cr;

    CacheVal
        *cache_val=NULL;

    int
        rc=0;

    char
        salt[9];

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL)
        return(-1);

    if (s_srv_conf->use_cache == 0)
        return(-1);

    if (r->user == NULL)
        return(-1);

    user=r->user;
    if (user == NULL)
        return(-1);

    m_rerr(r,cr->debug,"%s (%d) - mod_auth_ldap_insert_user_in_memory_cache() user: %s",MFL,user);

    /* insert the user data to hash table */
    cache_val=(CacheVal *) malloc(sizeof(CacheVal));
    if (cache_val == (CacheVal *) NULL)
    {
        m_rerr(r,1,"%s (%d) - Memrory allocation failed for Cache",MFL);
        rc=(-1);
        goto ExitProcessing;
    }
    memset(cache_val,0,sizeof(CacheVal));

    /* populate our cache data */

    /* ldap server */
    ap_snprintf(cache_val->ldap_server,sizeof(cache_val->ldap_server)-1,"%s",cr->ldap_server);

    /* ldap port */
    cache_val->ldap_port=cr->ldap_port;

    /* remote ip */
    ap_snprintf(cache_val->remote_ip,sizeof(cache_val->remote_ip)-1,"%s",r->connection->remote_ip);

    /* save user dn */
    if (cr->user_dn != NULL)
        ap_snprintf(cache_val->user_dn,sizeof(cache_val->user_dn)-1,"%s",cr->user_dn);
    else
    {
        m_rerr(r,cr->debug,"%s (%d) - Could not store user DN in cache, because use DN in cr is empty",MFL);
    }

    /* env var LDAP_USER_ALT */
    if (ldap_user_alt && *ldap_user_alt && cr->uid_attr_alt)
    {
        ap_snprintf(cache_val->ldap_user_alt,sizeof(cache_val->ldap_user_alt)-1,"%s",ldap_user_alt);
    }

    /* md5 crypt the password */
    mod_auth_ldap_to_64(&salt[0],rand(),8);
    salt[8]='\0';

    apr_md5_encode(sent_pw,(const char *) salt,cache_val->cpw,sizeof(cache_val->cpw));

    /* time */
    cache_val->t=time(NULL);
    cache_val->pid=getpid();

    if (cr->use_ght_hashlib)
    {
        if (ght_insert(s_srv_conf->hashp,cache_val,strlen(user),user) != 0)
        {
            m_rerr(r,cr->debug,"%s (%d) - Could not insert %s to hash errno: %d (%s)",MFL,user,errno,strerror(errno));
            rc=(-1);
            goto ExitProcessing;
        }
        rc=0; /* OK */
    }

    if (cr->debug)
    {
        m_rerr(r,cr->debug,"%s (%d) - INSERTED TO CACHE:",MFL);
        if (cr->use_ght_hashlib)
            m_rerr(r,cr->debug,"(insert)         Hash: %x, PID: %d",s_srv_conf->hashp,cache_val->pid);

        m_rerr(r,cr->debug,"(insert)  ldap_server: %s",cache_val->ldap_server);
        m_rerr(r,cr->debug,"(insert)    ldap_port: %d",cache_val->ldap_port);
        m_rerr(r,cr->debug,"(insert)    remote IP: %s",cache_val->remote_ip);
        m_rerr(r,cr->debug,"(insert)      user DN: %s",cache_val->user_dn);
        m_rerr(r,cr->debug,"(insert)         pass: <not shown>");
        m_rerr(r,cr->debug,"(insert) Last Authenticated: %s",mod_auth_ldap_ctime(&cache_val->t));
    }
ExitProcessing:

    if (rc == -1)
    {
        if (cache_val)
        {
            (void) free((char *) cache_val);
        }
    }
    return(rc);
}
/**
 * delete memory cache related to a user
 * returns 0 on success, -1 on failure
 */
static int mod_auth_ldap_delete_user_from_memory_cache(request_rec *r)
{
    CacheVal
        *cache_val=NULL;

    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (r->user == NULL)
        return(-1);

    if (cr->use_ght_hashlib)
    {
        cache_val=(CacheVal *) ght_remove(s_srv_conf->hashp,strlen(r->user),r->user);
        if (cache_val)
        {
            m_rerr(r,cr->debug,"%s (%d) - Deleted user %s from memory cache",MFL,r->user);
            (void) free((char *) cache_val);
        }
        else
        {
            m_rerr(r,cr->debug,"%s (%d) - Key not found for user %s to delete from cahce",
                MFL,r->user);
            return(-1);
        }
    }

    return(0);
}


#else
static int mod_auth_ldap_find_user_in_memory_cache(request_rec *r,const char *sent_pw)
{
    return(0);
}
static int mod_auth_ldap_insert_user_in_memory_cache(request_rec *r,const char *sent_pw,char *uid_attr_alt)
{
    return(0);
}
static int mod_auth_ldap_delete_user_from_memory_cache(request_rec *r)
{
    return(0);
}

#endif /* !USE_MEM_CACHE */


#ifdef USE_DATABASE
/*
 * Check fast disk based db to see if the user is there
 */
static int mod_auth_ldap_find_user_in_disk_cache(request_rec *r,const char *sent_pw)
{
    char
        *user=NULL;

    ldap_auth_config_rec
        *cr;

    CacheVal
        *cache_val;

    time_t
        timediff,
        os_now;

    int
        rc=1;

    LDBM
        dbp=NULL;

    Datum
        key_data,
        return_data;

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL)
        return(0);

    /* open the db */
    dbp=ldbm_open(s_srv_conf->cache_db_path,LDBM_WRCREAT,00600,0);
    if (dbp == NULL)
    {
        m_rerr(r,1,"%s (%d) Could not open cache db: %s",MFL,s_srv_conf->cache_db_path);
        rc=0;
        goto ExitProcessing2;
        return(0);
    }
    /* initialize datum */
    ldbm_datum_init(key_data);
    ldbm_datum_init(return_data);

    user=r->user;
    if (user == NULL)
        return(0);

    /* check for cache */
    key_data.dptr=user;
    key_data.dsize=strlen(user);

    return_data=ldbm_fetch(dbp,key_data);
    if (return_data.dptr == NULL)
    {
        m_rerr(r,cr->debug,"%s (%d) user %s is not in disk cache",MFL,user);
        rc=0;
        goto ExitProcessing2;
    }
    cache_val=(CacheVal *) return_data.dptr;
    if (cr->debug)
    {
        m_rerr(r,cr->debug,"%s (%d) - Found in Cache:",MFL);
        m_rerr(r,cr->debug,"(found)              PID: %d",cache_val->pid);
        m_rerr(r,cr->debug,"(found)  ldap_server: %s",cache_val->ldap_server);
        m_rerr(r,cr->debug,"(found)    ldap_port: %d",cache_val->ldap_port);
        m_rerr(r,cr->debug,"(found)    remote IP: %s",cache_val->remote_ip);
        m_rerr(r,cr->debug,"(found)      user DN: %s",cache_val->user_dn);
        /*m_rerr(r,cr->debug,"(found)         pass: %s",cache_val->cpw);*/
        m_rerr(r,cr->debug,"(found)         pass: <not shown>");
        m_rerr(r,cr->debug,"(found) Last Authenticated: %s",mod_auth_ldap_ctime(&cache_val->t));
    }

    /* Verify with data from connection */

    if (ap_strcasecmp_match(cr->ldap_server,cache_val->ldap_server) != 0)
    {
        /* ldap server not the same */
        m_rerr(r,cr->debug,"%s (%d) - Cache: ldap servers not the same, conn: %s, cache: %s",MFL,cr->ldap_server,cache_val->ldap_server);
        rc=0;
        goto ExitProcessing;
    }
    m_rerr(r,cr->debug,"%s (%d) - ldap server matched with cached one, conn: %s, cahce: %s",MFL,cr->ldap_server,cache_val->ldap_server);

    if (cr->ldap_port != cache_val->ldap_port)
    {
        m_rerr(r,cr->debug,"%s (%d) - Cache: ldap port not the same, conn: %d, cache: %d",MFL,cr->ldap_port,cache_val->ldap_port);
        /* ldap port not the same */
        rc=0;
        goto ExitProcessing;
    }
    m_rerr(r,cr->debug,"%s (%d) - ldap port matched with cached one, conn: %d, cache: %d",MFL,cr->ldap_port,cache_val->ldap_port);

    if (ap_strcasecmp_match(r->connection->remote_ip,cache_val->remote_ip) != 0)
    {
        /* connected from different machine */
        m_rerr(r,cr->debug,"%s (%d) - Cache: remote ip did not match, conn: %s, cahce: %s",
                MFL,r->connection->remote_ip,cache_val->remote_ip);
        rc=0;
        goto ExitProcessing;
    }

    m_rerr(r,cr->debug,"%s (%d) - remote IP matched with cached one, conn: %s, cahce: %s",
            MFL,r->connection->remote_ip,cache_val->remote_ip);


    if (apr_password_validate(sent_pw,cache_val->cpw) != 0)
    {
        m_rerr(r,cr->debug,"%s (%d) - Cache: passwords do no match",MFL);
        /* password does not match */
        rc=0;
        goto ExitProcessing;
    }
    m_rerr(r,cr->debug,"%s (%d) - password matched with cached one",MFL);

    os_now=time(NULL);
    timediff=os_now - cache_val->t;
    if (timediff > s_srv_conf->cache_ttl)
    {
        m_rerr(r,cr->debug,"%s (%d) - *** Cache expired (%d sec old), TTL=%d",MFL,timediff,s_srv_conf->cache_ttl);
        m_rerr(r,cr->debug,"%s (%d) - Cache last inserted: %s",MFL,mod_auth_ldap_ctime(&cache_val->t));
        /* autheticated too long ago */
        rc=0;
        goto ExitProcessing;
    }
    /* set env vars */
    ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",MOD_AUTH_LDAP_VERSION_S);
    ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_CACHE_TYPE","Disk");
    ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
    if (*cache_val->ldap_user_alt)
    {
        m_rerr(r,cr->debug,"%s (%d) - Setting LDAP_USER_ALT from disk cache: %s",MFL,cache_val->ldap_user_alt);
        ap_table_setn(r->subprocess_env,"LDAP_USER_ALT",cache_val->ldap_user_alt);
    }


    m_rerr(r,cr->debug,"%s (%d) - Cache not expired yet (%d sec old), TTL=%d",MFL,timediff,s_srv_conf->cache_ttl);
    rc=1;


ExitProcessing:

    if (rc == 0)
    {
        if (ldbm_delete(dbp,key_data) == 0)
        {
            m_rerr(r,cr->debug,"%s (%d) - Deleted user %s from disk cache",MFL,user);
        }
        else
        {
            m_rerr(r,cr->debug,"%s (%d) - Could not delete %s from disk cache",
                    MFL,user);
        }
    }

ExitProcessing2:

    if (return_data.dptr)
        ldbm_datum_free(dbp,return_data);

    if (dbp)
        ldbm_close(dbp);

    return(rc);
}

/*
 * Insert user data to disk cache. This routine is called after the user is authenticated
 */
static int mod_auth_ldap_insert_user_in_disk_cache(request_rec *r,const char *sent_pw,char *ldap_user_alt)
{
    char
        *user=NULL;

    ldap_auth_config_rec
        *cr;

    CacheVal
        *cache_val;

    char
        salt[9];

    int
        data_len=0,
        rc=0;

    LDBM
        dbp=NULL;

    Datum
        key_data,
        data_data;


    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL)
        return(0);

    if (s_srv_conf->use_cache == 0)
        return(0);

    m_rerr(r,cr->debug,"%s (%d) - Cache Db: %s",MFL,s_srv_conf->cache_db_path);
    /* open the database */
    dbp=ldbm_open(s_srv_conf->cache_db_path,LDBM_WRCREAT,00600,0);
    if (dbp == NULL)
    {
        m_rerr(r,cr->debug,"%s (%d) - could not open dbb for writing: %s",MFL,s_srv_conf->cache_db_path);
        return(0);
    }

    /* initialize datum */
    ldbm_datum_init(key_data);
    ldbm_datum_init(data_data);

    user=r->user;
    if (user == NULL)
    {
        rc=0;
        goto ExitProcessing;
    }
    m_rerr(r,cr->debug,"%s (%d) - mod_auth_ldap_insert_user_in_memory_cache() user: %s",MFL,user);

    /* insert the user data to hash table */
    cache_val=(CacheVal *) malloc(sizeof(CacheVal));
    if (cache_val == (CacheVal *) NULL)
    {
        m_rerr(r,1,"%s (%d) - Memrory allocation failed for Cache",MFL);
        return(0);
    }
    memset(cache_val,0,sizeof(CacheVal));

    /* populate our cache data */

    /* ldap server */
    ap_snprintf(cache_val->ldap_server,sizeof(cache_val->ldap_server)-1,"%s",cr->ldap_server);
    data_len += strlen(cache_val->ldap_server);

    /* ldap port */
    cache_val->ldap_port=cr->ldap_port;
    data_len += 4;

    /* save user dn */
    if (cr->user_dn != NULL)
        ap_snprintf(cache_val->user_dn,sizeof(cache_val->user_dn)-1,"%s",cr->user_dn);
    else
    {
        m_rerr(r,cr->debug,"%s (%d) - Could not store user DN in cache, because use DN in cr is empty",MFL);
    }

    /* remote ip */
    ap_snprintf(cache_val->remote_ip,sizeof(cache_val->remote_ip)-1,"%s",r->connection->remote_ip);
    data_len += strlen(cache_val->remote_ip);

    /* md5 crypt the password */
    mod_auth_ldap_to_64(&salt[0],rand(),8);
    salt[8]='\0';

    apr_md5_encode(sent_pw,(const char *) salt,cache_val->cpw,sizeof(cache_val->cpw));
    data_len += strlen(cache_val->cpw);

    /* time */
    /*
    cache_val->t=apr_time_now();
    */
    cache_val->t=time(NULL);
    data_len += 4;
    cache_val->pid=getpid();
    data_len += 4;

    key_data.dptr=user;
    key_data.dsize=strlen(user);

    data_data.dptr=cache_val;
    m_rerr(r,cr->debug,"%s (%d) key len: %d",MFL,strlen(user));
    data_data.dsize=sizeof(CacheVal);
    m_rerr(r,cr->debug,"%s (%d) data len: %d",MFL,data_len);

    if (ldbm_store(dbp,key_data,data_data,LDBM_REPLACE) != 0)
    {
        m_rerr(r,1,"%s (%d) - could not save cache data for user: %s (%s)",MFL,user,strerror(errno));
        rc=0;
        goto ExitProcessing;
    }

    if (cr->debug)
    {
        m_rerr(r,cr->debug,"%s (%d) - INSERTED TO DISK CACHE:",MFL);
        m_rerr(r,cr->debug,"(insert)          PID: %d",cache_val->pid);
        m_rerr(r,cr->debug,"(insert)  ldap_server: %s",cache_val->ldap_server);
        m_rerr(r,cr->debug,"(insert)    ldap_port: %d",cache_val->ldap_port);
        m_rerr(r,cr->debug,"(insert)    remote IP: %s",cache_val->remote_ip);
        m_rerr(r,cr->debug,"(insert)      user DN: %s",cache_val->user_dn);
        /*m_rerr(r,cr->debug,"(insert)         pass: %s",cache_val->cpw);*/
        m_rerr(r,cr->debug,"(insert)         pass: <not shown>");
        m_rerr(r,cr->debug,"(insert) Last Authenticated: %s",mod_auth_ldap_ctime(&cache_val->t));
 
    }
ExitProcessing:

    if (dbp)
        ldbm_close(dbp);

    if (cache_val)
        (void) free((char *) cache_val);

    return(rc);
}

static int mod_auth_ldap_delete_user_from_disk_cache(request_rec *r)
{
    int
        rc=0;

    return(0);
}

#endif /* USE_DATABASE */


/**
 * delete user from memory or disk cache
 * returns 0 on success -1 on failure
 */
static int mod_auth_ldap_delete_user_from_cache(request_rec *r)
{
#ifdef USE_MEM_CACHE
    return(mod_auth_ldap_delete_user_from_memory_cache(r));
#endif /* USE_MEM_CACHE */

#ifdef USE_DATABASE
    return(mod_auth_ldap_delete_user_from_disk_cache(r));
#endif /* USE_DATABASE */

    return(0);
}


/*
** remove all whitespace (leading or trailing) from a string.
** borrowed from public domain c snippets library.
** muquit, sep-07-1998
*/
static char *rmallws(char *str)
{
    char
    *obuf,
    *nbuf;

    if (str)
    {
        for (obuf=str, nbuf=str; *obuf; ++obuf)
        {
            if (!isspace(*obuf))
                *nbuf++ = *obuf;
        }
        *nbuf='\0';
    }
    return (str);
}

/* for my use only - muquit@muquit.com */
#ifdef MY_DEBUG
static void dumpResult(LDAP *ld,LDAPMessage *res,request_rec *r)
{
    LDAPMessage
    *ent;

    BerElement
    *ber;

    int
    i;

    char
    *attr,
    **vals;

    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "XXXXXXXX");
    for (ent=ldap_first_entry
             (ld,res); ent != NULL; ent=ldap_next_entry
                                        (ld,ent))
    {
        for (attr=ldap_first_attribute(ld,ent,&ber); attr != NULL;
                attr=ldap_next_attribute(ld,ent,ber))
        {
            if (strcasecmp(attr,"jpegphoto") != 0)
            {
                if ((vals=ldap_get_values(ld,ent,attr)) != NULL)
                {
                    for (i=0; vals[i] != NULL; i++)
                    {
                        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                                      "dump: %s=%s",attr,vals[i]);
                    }
                    if (vals)
                        ldap_value_free(vals);
                }
            }
        }
    }
}
#endif /* MY_DEBUG */

/*
 * search cr->uid_attr_alt and set the value to LDAP_USE_ALT env var
 */
static char *mod_auth_ldap_set_ldap_user_alt(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        rc;

    char
        *v=NULL,
        **vals=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;


    cr= (ldap_auth_config_rec *)
        ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    if (cr == NULL)
        return;

    if (cr->ld == NULL)
        return;

    (void) ap_snprintf(cr->filter,sizeof(cr->filter)-1,"(%s=*)",
                       cr->uid_attr_alt);
    rc=ldap_search_s(cr->ld,
            cr->user_dn,
            LDAP_SCOPE_SUBTREE,
            cr->filter,
            NULL,
            0,
            &res);

    if (rc != LDAP_SUCCESS)
    {
        m_rerr(r,cr->debug,
                "%s (%d) - search failed with filter: %s",MFL,cr->filter);
        return;
    }

    /* we;ll set the value of the first match */
    ent=ldap_first_entry(cr->ld,res);
    if (ent)
    {
        vals=ldap_get_values(cr->ld,ent,cr->uid_attr_alt);
        if (vals != NULL)
        {
            /*
            m_rerr(r,cr->debug,"%s (%d) - %d - %s=%s",MFL,cr->uid_attr_alt,val[0]);
            */
            /* XXX: need to free memory? */
            v=ap_pstrdup(r->pool,vals[0]);
            m_rerr(r,cr->debug,"%s (%d) - Setting LDAP_USER_ALT %s=%s",MFL,cr->uid_attr_alt,v);
            ap_table_setn(r->subprocess_env,"LDAP_USER_ALT",v);
            ldap_value_free(vals);
        }
    }


    if (res != NULL)
        ldap_msgfree(res);

    return(v);
}

/*
** Initilize ld 
** returns pointer to LDAP on success, NULL on failure
*/
LDAP *mod_auth_ldap_initialize_ldap(request_rec *r)
{
    LDAP
        *ld=NULL;

    int
        rc,
        timeout=0,
        ldap_vers;

    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    int
        err=0,
        ssl_ok=0;

    /* If a previsous connection is made from that pid or thread, we'll re-use it */
    if (! cr->start_ssl) /* we'll set ldapv3 if ssl is on */
    {
        m_rerr(r,cr->debug,"%s (%d) - Setting LDAP protocol version to: %d",MFL,cr->ldap_protocol_version);

        /* Set LDAP version to whatever configured */
        ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION,&cr->ldap_protocol_version);

        /*  Set alias deference option*/
        if (ldap_set_option(NULL,LDAP_OPT_DEREF,&cr->ldap_opt_deref) != LDAP_SUCCESS)
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"%s (%d) - failed setting LDAP_DEREF",MFL);
        }
    }

#ifdef HAVE_SSL
    if (cr->start_ssl && cr->ca_cert_db_path) /* cert7.db path specified */
    {
        /* initialize the SSL client */
        if (ldapssl_client_init(cr->ca_cert_db_path,NULL) != 0)
        {
            m_rerr(r,1,"%s (%d) - Could not initialize SSL client, cert db path: %s",MFL,cr->ca_cert_db_path);
            err++;
            goto ExitProcessing;
        }

        m_rerr(r,cr->debug,"%s (%d) - SSL client initialized, certificate db dir %s is valid",
                MFL,cr->ca_cert_db_path);

        /* get a handle to LDAP connection */
        if (s_srv_conf->ld == NULL)
        {
                m_rerr(r,cr->debug,
                    "%s (%d) - INITIALIZING NEW SSL CONNECTION",MFL);
                s_srv_conf->ld=ldapssl_init(cr->ldap_server,cr->ldap_port,1);
                ld=s_srv_conf->ld;
        }
        else
        {
            ld=s_srv_conf->ld;
            m_rerr(r,cr->debug,"%s (%d) - SSL using already open LDAP connection: 0x%x",MFL,ld);
        }

        if (ld == NULL)
        {
            m_rerr(r,1,"%s (%d) - Could not initialize SSL connection, LDAP server %s at port %d",MFL,cr->ldap_server,cr->ldap_port);
            err++;
            goto ExitProcessing;
        }
   
        m_rerr(r,cr->debug,"%s (%d) - Initialized SSL connection to LDAP server %s:%d with SSL",MFL,cr->ldap_server,cr->ldap_port);


        /* use LDAPv3 */
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - SSL ldap_set_option(LDAP_VERSION3) failed (%d): %s",rc,ldap_err2string(rc));
            err++;
            goto ExitProcessing;
        }
        ssl_ok=1;
    }
#endif /* HAVE_SSL */

    /* if we already didn't initialize ssl connection */
    if (ssl_ok == 0)
    {
        if (ld == NULL)
        {
            m_rerr(r,cr->debug,"%s (%d) - ** INITIALIZING NEW LDAP CONNECTION",MFL);
            s_srv_conf->ld=ldap_init(cr->ldap_server,cr->ldap_port);
            ld=s_srv_conf->ld;

            m_rerr(r,cr->debug,"%s (%d) - ld: 0x%x",MFL,ld);
        }
        else
        {
            ld=s_srv_conf->ld;
            m_rerr(r,cr->debug,"%s (%d) - ** USING ALREADY OPEN LDAP CONNECTION: %x",MFL,ld);
        }
    }

    /* set connect timeout.. iPlanet SDK at this time */
#if defined(LDAP_X_OPT_CONNECT_TIMEOUT)
        timeout=cr->ldap_connect_timeout*1000;
        m_rerr(r,cr->debug,
                "%s (%d) - Setting connect timeout to: %d seconds",MFL,cr->ldap_connect_timeout);

        rc=ldap_set_option(ld,LDAP_X_OPT_CONNECT_TIMEOUT,&timeout);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,
                "%s (%d) - Could not set LDAP_X_OPT_CONNECT_TIMEOUT (%s)",MFL,ldap_err2string(rc));
                                
        }
        else
        {
            m_rerr(r,cr->debug,
                "%s (%d)- Successfully set connection timeout to %d seconds",MFL,cr->ldap_connect_timeout);
        }
#else
        m_rerr(r,cr->debug,"%s (%d) - not compiled with iPlanet C SDK, connect timeout will not be available",MFL);
#endif /* LDAP_X_OPT_CONNECT_TIMEOUT */

    /*
    ** so far openldap sdk and server supports STARTTLS to encrypt
    ** traffic. muquit@muquit.com Dec-24-2002 
    */
#ifdef HAVE_LDAP_START_TLS_S
    if (cr->auth_ldapstarttls == 1)
    {
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - ldap_set_option(LDAP_VERSION3) failed (%d): %s",rc,ldap_err2string(rc));
        }

        rc=ldap_start_tls_s(ld,NULL,NULL);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"%s (%d) - Starting TLS failed (%d): %s",MFL,rc,ldap_err2string(rc));
            err++;
            goto ExitProcessing;
        }
    }
#endif /* HAVE_LDAP_START_TLS_S */


ExitProcessing:
    if (err)
    {
        ldap_unbind(ld); /* we don't care if it fails */
        ld=NULL;
    }
    if (ld)
        cr->ld=ld;
    return(ld);
}

/*
** validate the user id via LDAP
** Returns:
**      DECLINED        if ldap server is not defined in config file
**      HTTP_UNAUTHORIZED   for any other kind of errors
**      OK              if authorized
*/
static int mod_auth_ldap_check_user_authentication(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        from_cache=0,
        rcode=OK,
        rc,
        ldap_vers;

    int
        timeout=0,
        ssl_ok=0,
        res;

    conn_rec
        *c;

    const char
        *sent_pw=NULL;

    char
        *ldap_user_alt=NULL,
        *dn=NULL;

    int
        try_again=1,
        bind_try=1;

    char
      authonbind_dn[256];

    LDAP
        *ld=NULL;

#ifdef USE_MEM_CACHE
#define MEM_CACHE_SUPPORT  "With Memory Cache support"
#else
#define MEM_CACHE_SUPPORT  "Without Memory Cache support"
#endif /* USE_MEM_CACHE */

#ifdef USE_DATABASE
#define DISK_CACHE_SUPPORT "With Fast disk Cache support"
#else
#define DISK_CACHE_SUPPORT "Without Disk Cache support"
#endif /* USE_DATABASE */

#ifdef HAVE_SSL
#define SSL_SUPPORT       "With SSL Support"
#else
#define SSL_SUPPORT       "Without SSL Support"
#endif /* HAVE_SSL */

#ifdef HAVE_LDAP_START_TLS_S
#define START_TLS_SUPPORT   "With OpenLDAP Start TLS"
#else
#define START_TLS_SUPPORT   "Without OpenLDAP Start TLS"
#endif /* HAVE_LDAP_START_TLS_S */

    ssl_ok=0;
    c=r->connection;
    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    res=ap_get_basic_auth_pw(r,&sent_pw);
    if (res)
        return (res);

    /* if no user is specified, keep asking */
    if (r->user == NULL || *r->user == '\0')
    {
        m_rerr(r,cr->debug,"%s (%d) - This LDAP auth module does not accept empty user",MFL);
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (HTTP_UNAUTHORIZED);
        }
        else
            return(DECLINED);
    }

    /* if no password is specified, keep asking */
    if ((sent_pw == NULL) || (*sent_pw == '\0'))
    {
        m_rerr(r,cr->debug,"%s (%d) - This LDAP auth module does not accept empty password",MFL);
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (HTTP_UNAUTHORIZED);
        }
        else
            return(DECLINED);
    }

    /*
        m_rerr(r,cr->debug,"[%s (%d)] - MMM Auth name: %s",MFL,ap_auth_name(r));
        m_rerr(r,cr->debug,"[%s (%d)] - MMM Auth type: %s",MFL,ap_auth_type(r));
    */
        /* -- Lock starts -- */

    TMPM_LOCK;


    m_rerr(r,cr->debug,"%s (%d) - mod_auth_ldap v%s Compiled: %s, %s, %s, %s, URL: %s",
            MFL,MOD_AUTH_LDAP_VERSION_S,
            MEM_CACHE_SUPPORT,
            DISK_CACHE_SUPPORT,
            SSL_SUPPORT,
            START_TLS_SUPPORT,
            MOD_AUTH_LDAP_SITE_URL);

    if (cr->ldap_server == NULL)
    {
        m_rerr(r,1,"%s (%d) - No LDAP server specified",MFL);
        rcode=DECLINED;
        goto exit_processing;
    }

    if (cr->ldap_server)
    {
        m_rerr(r,cr->debug,"%s (%d) - LDAP server=%s,Port=%d",
            MFL,cr->ldap_server,cr->ldap_port);
    }

    /*
     * check if caching is on first. make sure either of meory or disk caching is on.
     * obviously we will not use both method at the same time. If both methohds are
     * on, turn off caching
     */

     m_rerr(r,cr->debug,"%s (%d) - s_srv_conf->use_memory_cache=%d,s_srv_conf->use_disk_cache=%d",
             MFL,s_srv_conf->use_memory_cache,s_srv_conf->use_disk_cache);
    if (s_srv_conf->use_memory_cache && s_srv_conf->use_disk_cache)
    {
        m_rerr(r,cr->debug,"%s (%d) - LDAP_UseMemoryCache_C and LDAP_UseDiskCache_G can not be used together, caching will be disabled",
                MFL);
        s_srv_conf->use_cache=0;
        s_srv_conf->use_memory_cache=0;
        s_srv_conf->use_disk_cache=0;
    }
    else
    {
        if (s_srv_conf->use_memory_cache)
        {
            s_srv_conf->use_cache=1;
        }

        /* if using memory cache the following condition willbe false and vice versa */
        if (s_srv_conf->use_disk_cache)
        {
#ifdef USE_DATABASE
            m_rerr(r,cr->debug,"%s (%d) - Disk Caching is ON",MFL);
            s_srv_conf->use_cache=1;
#else
            m_rerr(r,cr->debug,"%s (%d) - Disk Caching will be OFF, not compiled with database support",MFL);
#endif /* USE_DATABASE */
        }
    }

#if defined(USE_MEM_CACHE)
    if (s_srv_conf->use_cache && s_srv_conf->use_memory_cache)
    {
        if (cr->use_ght_hashlib)
        {
            if (s_srv_conf->hashp == NULL)
            {
                m_rerr(r,cr->debug,"%s (%d) - Creating hash table of size 1024",MFL);
                s_srv_conf->hashp=ght_create(1024);
            }
        }
    }
#endif /* USE_MEM_CACHE */


    if (! cr->start_ssl) /* we'll set ldapv3 if ssl is on */
    {
        m_rerr(r,cr->debug,"%s (%d) - Setting LDAP protocol version to: %d",
                MFL,cr->ldap_protocol_version);

        /* Set LDAP version to whatever configured */
        ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION,
                &cr->ldap_protocol_version);

        /*  Set alias deference option*/
        if (ldap_set_option(NULL,LDAP_OPT_DEREF,&cr->ldap_opt_deref) 
                != LDAP_SUCCESS)
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                          "[mod_auth_ldap.c] - failed setting LDAP_DEREF");
        }
    }


    /* check cache first */
    if (s_srv_conf->use_cache && s_srv_conf->use_memory_cache)
    {

#if defined(USE_MEM_CACHE)
        rc=mod_auth_ldap_find_user_in_memory_cache(r,sent_pw);
        m_rerr(r,cr->debug,"%s (%d) - rc from mod_auth_ldap_find_user_in_memory_cache=%d",MFL,rc);
        if (rc == 1)
        {
            TMPM_UNLOCK;
            
            m_rerr(r,cr->debug,"%s (%d) - user %s is Authenticated (from Memory Cache)",MFL,r->user);

            m_log_cache_stats(r,cr->log_cache_stats,
                        "%s (%d) - Authentication %d %d",MFL,ght_tablesize(s_srv_conf->hashp),ght_size(s_srv_conf->hashp));
            return(OK);
        }
#endif /*  USE_MEM_CACHE */

    } /* use_cache */

    if (s_srv_conf->use_cache && s_srv_conf->use_disk_cache)
    {
#ifdef USE_DATABASE
        if (s_srv_conf->cache_db_path)
        {
            m_rerr(r,cr->debug,"%s (%d) - Cache Db: %s",MFL,s_srv_conf->cache_db_path);
            rc=mod_auth_ldap_find_user_in_disk_cache(r,sent_pw);
            if (rc == 1)
            {
                TMPM_UNLOCK;
                m_rerr(r,cr->debug,"%s (%d) - user %s is Authenticated (from fast disk Cache)",MFL,r->user);
                return(OK);
            }
        }
#endif /* USE_DATABASE */

    }

    /*
    ** If a previsous connection is made from that pid or thread, we'll 
    ** re-use it
    */
    bind_try=1;
try_again:

#ifdef HAVE_SSL
    if (cr->start_ssl && cr->ca_cert_db_path) /* cert7.db path specified */
    {
        /* initialize the SSL client */
        if (ldapssl_client_init(cr->ca_cert_db_path,NULL) != 0)
        {
            m_rerr(r,1,"%s (%d) - Could not initialize SSL client, cert db path: %s",MFL,cr->ca_cert_db_path);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
        m_rerr(r,cr->debug,"%s (%d) - SSL client initialized, certificate db dir %s is valid",
                MFL,cr->ca_cert_db_path);

        /* get a handle to LDAP connection */
        if (s_srv_conf->ld == NULL)
        {
                m_rerr(r,cr->debug,
                    "%s (%d) - INITIALIZAING NEW SSL LDAP CONNECTION, attempt# %d",MFL,bind_try);
                s_srv_conf->ld=ldapssl_init(cr->ldap_server,cr->ldap_port,1);
                cr->ld=s_srv_conf->ld;
        }
        else
        {
            cr->ld=s_srv_conf->ld;
            m_rerr(r,cr->debug,"%s (%d) - SSL using already open ldap connection: 0x%x",
                    MFL,cr->ld);
        }

        if (cr->ld == NULL)
        {
            m_rerr(r,1,"%s (%d) - Could not initialize SSL connection, LDAP server %s at port %d",MFL,cr->ldap_server,cr->ldap_port);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
   
        m_rerr(r,cr->debug,"%s (%d) - Initialized SSL connection to LDAP server %s:%d with SSL",
                MFL,cr->ldap_server,cr->ldap_port);


        /* use LDAPv3 */
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(cr->ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - SSL ldap_set_option(LDAP_VERSION3) failed (%d): %s",rc,ldap_err2string(rc));
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
        ssl_ok=1;
    }
#endif /* HAVE_SSL */

    /* connect to LDAP server if we already didn't make ssl connection*/
    if (ssl_ok == 0)
    {
        /*
        if (cr->ld == NULL)
        */
        if (s_srv_conf->ld == NULL)
        {
            m_rerr(r,cr->debug,"%s (%d) - ** INITIALIZING NEW LDAP CONNECTION, try# %d",MFL,bind_try);
                s_srv_conf->ld=ldap_init(cr->ldap_server,cr->ldap_port);
                cr->ld=s_srv_conf->ld;

            m_rerr(r,cr->debug,"%s (%d) - cr->ld: 0x%x",MFL,cr->ld);
        }
        else
        {
            cr->ld=s_srv_conf->ld;
            m_rerr(r,cr->debug,
                "%s (%d) - ** USING ALREADY OPEN LDAP CONNECTION: %x",MFL,cr->ld);
        }

        if (cr->ld == (LDAP *) NULL)
        {
            m_rerr(r,1,"%s (%d) - Could not initialize connection to LDAP server %s at port %d",MFL,cr->ldap_server,cr->ldap_port);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
    }

    /* set connect timeout.. iPlanet SDK at this time */
#if defined(LDAP_X_OPT_CONNECT_TIMEOUT)
        timeout=cr->ldap_connect_timeout*1000;
        m_rerr(r,cr->debug,
                "%s (%d) - Setting connect timeout to: %d seconds",MFL,cr->ldap_connect_timeout);

        rc=ldap_set_option(cr->ld,LDAP_X_OPT_CONNECT_TIMEOUT,
                &timeout);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,
                "%s (%d) - Could not set LDAP_X_OPT_CONNECT_TIMEOUT (%s)",
                MFL,ldap_err2string(rc));
                                
        }
        else
        {
            m_rerr(r,cr->debug,
                "%s (%d) - Successfully set connection timeout to %d seconds",
                MFL,cr->ldap_connect_timeout);
        }
#else
        m_rerr(r,cr->debug,"%s (%d) - you didn't compile with iPlanet C SDK, connect timeout will not be available",MFL);
#endif /* LDAP_X_OPT_CONNECT_TIMEOUT */

    /*
    ** so far openldap sdk and server supports STARTTLS to encrypt
    ** traffic. muquit@muquit.com Dec-24-2002 
    */
#ifdef HAVE_LDAP_START_TLS_S
    if (cr->auth_ldapstarttls == 1)
    {
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(cr->ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"%s (%d) - ldap_set_option(LDAP_VERSION3) failed (%d): %s",MFL,rc,ldap_err2string(rc));
        }

        rc=ldap_start_tls_s(cr->ld,NULL,NULL);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"%s (%d) - Starting TLS failed (%d): %s",MFL,rc,ldap_err2string(rc));
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
    }
#endif /* HAVE_LDAP_START_TLS_S */


    if (!(cr->auth_on_bind))
    {
        /* now get the User DN */
        dn=mod_auth_ldap_find_user_dn(r,&try_again);

        if (dn == (char *) NULL)
        {
          m_rerr(r,cr->debug,"%s (%d) - Bind attempt# %d, cound not find DN for user \"%s\" with attr \"%s\"",MFL,bind_try,r->user,cr->uid_attr);

          /*
          ** It's possible already open connection by the module is closed 
          ** by the ldap server or ldap server is down. Some people 
          ** reported this kind of problem.  muquit@muquit.com Jun-09-2003 
          */
          if (try_again && (bind_try < cr->max_bind_tries))
          {
              ldap_unbind(cr->ld); /* we don't care if it fails */
              cr->ld=NULL;
              s_srv_conf->ld=NULL;
              bind_try++;
              goto try_again;
          }
#if 0
          if (!try_again)
          {
              m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Connection to LDAP server is open but user does not exist, will not retry",MML);
          }
#endif /* if 0 */

          /* pass control to lower modules if AuthLDAPAuthoritative=no */
          if (! (cr->auth_ldapauthoritative))
          {
              m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

              rcode=DECLINED;
              goto exit_processing;
          }
          if (try_again)
          {
            m_rerr(r,cr->debug,"%s (%d) - Tried to bind %d times. Giving up.",MFL,bind_try);
          }
          rcode=HTTP_UNAUTHORIZED;
          goto exit_processing;
        }
    }
    else
    {
        m_rerr(r,cr->debug,"%s (%d) -  AuthOnBind is On, skipping LDAP Search",MFL);

        /* Build our dn from Base_DN and the authenticating username */
        if (cr->sub_dn)
        {
            ap_snprintf(authonbind_dn,sizeof(authonbind_dn)-1,"%s=%s,%s,%s",
                cr->uid_attr,r->user,cr->sub_dn,cr->base_dn);
        }
        else
        {
            ap_snprintf(authonbind_dn,sizeof(authonbind_dn)-1,"%s=%s,%s",
                cr->uid_attr,r->user,cr->base_dn);
        }
        dn=authonbind_dn;
        m_rerr(r,cr->debug,"%s (%d) - AuthOnBind is On, Using explicit user DN \"%s\"",MFL,dn);
    }

    /* store user DN, we'll use it later if needed*/
    cr->user_dn=ap_pstrdup(r->pool,dn);

    m_rerr(r,cr->debug,"%s (%d) - Bind attempt# %d, User DN: %s",MFL,bind_try,cr->user_dn);

    /*
    ** now if we can successfully bind to LDAP server as the specified
    ** userid and passoword, the user is a valid one!
    */
    if (ldap_simple_bind_s(cr->ld,dn,sent_pw) != LDAP_SUCCESS)
    {
        if (r->user)
            m_rerr(r,cr->debug,"%s (%d) - Invalid Login or Password for: %s",
                    MFL,r->user);

        if (!(cr->auth_on_bind))
        {
            /* use found, probably failed due to wrong password , don't 
             * bother to try again
             */
            if (cr->user_dn)
            {
                if (! (cr->auth_ldapauthoritative))
                {
                    m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

                    rcode=DECLINED;
                    goto exit_processing;
                }
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }

        /*
         * we didn't do any search so we don't if it failed because binding failed
         * or because the user does not exist
         */
        if (bind_try < cr->max_bind_tries)
        {
            ldap_unbind(cr->ld); /* we don't care if it fails */
            cr->ld=NULL;
            s_srv_conf->ld=NULL;
            bind_try++;
            goto try_again;
        }

        m_rerr(r,cr->debug,"%s (%d) - Tried to bind %d times. Giving up!", MFL,bind_try);

        /* pass control to lower modules if AuthLDAPAuthoritative=no */
        if (! (cr->auth_ldapauthoritative))
        {
            m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

            rcode=DECLINED;
            goto exit_processing;
        }

        rcode=HTTP_UNAUTHORIZED;
        goto exit_processing;
    }


exit_processing:

    /* if caching is on, insert user data to cache */
    if (rcode == OK)
    {
        /* Set environment variables */
        ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
        ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",MOD_AUTH_LDAP_VERSION_S);

        /* won't set passwd in env */
        /*
        if (cr->set_pass_to_env && sent_pw)
        {
            ap_table_setn(r->subprocess_env,"LDAP_PASSWORD",sent_pw);
        }
        */

        /*
         * if UID_Attr_Use is set, search this attribute and set the
         * value to LDAP_USER_ALT env variable
         */
        if (cr->uid_attr_alt)
        {
            ldap_user_alt=mod_auth_ldap_set_ldap_user_alt(r);
        }
        cr->env_vars_set=1;

#if defined(USE_MEM_CACHE)
        /* insert the user data into cache */
        if (s_srv_conf->use_cache && s_srv_conf->use_memory_cache)
        {
            m_rerr(r,cr->debug,"%s (%d) - Inserting user data to memory cache",MFL);
            mod_auth_ldap_insert_user_in_memory_cache(r,sent_pw,ldap_user_alt);
        }
#endif /* USE_MEM_CACHE */

        if (s_srv_conf->use_cache && s_srv_conf->use_disk_cache)
        {
#ifdef USE_DATABASE
            /* insert the user data into disk cache */
            if (s_srv_conf->cache_db_path)
            {
                m_rerr(r,cr->debug,"%s (%d) - Inserting user data to disk cache",MFL);
                mod_auth_ldap_insert_user_in_disk_cache(r,sent_pw,ldap_user_alt);
            }
            else
            {
                m_rerr(r,cr->debug,"%s (%d) - No Cache db path specified, will not save user data to disk",MFL);
            }
#endif /* USE_DATABASE */
        }
    }


    /* free things we won't need */
    if (dn != (char *) NULL)
    {
        if (! cr->auth_on_bind)
        {
            mod_auth_ldap_free_dn(dn);
        }
        dn=NULL;
    }

    TMPM_UNLOCK;    

    if (rcode == OK)
    {
        m_rerr(r,cr->debug,"%s (%d) - user %s is Authenticated (from LDAP)",MFL,r->user);
    }

    /* If we've turned off persistent connections, close ours. */
    if (! s_srv_conf->ldap_persistent)
    {
      m_rerr(r,cr->debug,"%s (%d) - LDAP_Persistent is off.  Closing old connection",MFL);
      ldap_unbind(s_srv_conf->ld);
      s_srv_conf->ld = NULL;
    }

    if (rcode == HTTP_UNAUTHORIZED)
        ap_note_basic_auth_failure(r);
    return (rcode);
}


/* called to check to see if resource being requested requires authorization. */
static int mod_auth_ldap_check_user_authorization(request_rec *r)
{
    const apr_array_header_t
        *reqs_arr;

    require_line
        *reqs;

    register int
        x;

    int
        res,
        rcode=OK,
        m;

    const char
        *sent_pw=NULL,
        *t,
        *w;

    ldap_auth_config_rec
        *cr;

    conn_rec
        *c;

    char
        *user_dn=NULL,
        *attr,
        *user=r->user;

    int
        user_dn_differs=0,
        retrieved_user_dn=0,
        try_again=0;

    CacheVal
        *cache_val=NULL;

    cr=(ldap_auth_config_rec *) NULL;
    reqs_arr=(ap_array_header_t *)NULL;
    reqs=(require_line *) NULL;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    reqs_arr=ap_requires(r);
    reqs=reqs_arr ? (require_line *) reqs_arr->elts : NULL;
    m=r->method_number;
    c=r->connection;

    res=ap_get_basic_auth_pw(r,&sent_pw);

    TMPM_LOCK;

    /* mm, Aug-20-1999 */
    if ((!s_srv_conf->use_cache) && cr->ldap_server == NULL)
    {
        if (cr->auth_ldapauthoritative)
            rcode=HTTP_UNAUTHORIZED;
        else
            rcode=DECLINED;
        goto ExitProcessing;
    }

    if ((! s_srv_conf->use_cache) && cr->ld == (LDAP *) NULL)
    {
        m_rerr(r,cr->debug,"%s (%d) - no open LDAP connection",MFL);
        if (cr->auth_ldapauthoritative)
            rcode=HTTP_UNAUTHORIZED;
        else
            rcode=DECLINED;
        goto ExitProcessing;
    }

    /* server cfg file must have at least one require line */
    if (reqs_arr == NULL)
    {
        m_rerr(r,1,"[mod_auth_ldap.c] - no require line in config file");
        if (cr->auth_ldapauthoritative)
            rcode=HTTP_UNAUTHORIZED;
        else
            rcode=DECLINED;
        goto ExitProcessing;
    }

    for (x=0; x < reqs_arr->nelts; x++)
    {
        if (!(reqs[x].method_mask & (1 << m)))
            continue;
        t=reqs[x].requirement;

        /*
        ** use ap_getword_white() instead of ap_getword() and my own rmallws
        ** muquit@muquit.com, Feb-10-2000
        */
        w=ap_getword_white(r->pool,&t);

        /*
        ** if "require valid-user" found in cfg, the user is authorized.
        ** we'r outta here with OK
        */
        if(strcmp(w,"valid-user") == 0)
        {
            rcode=OK;
            goto ExitProcessing;
        }
        else if (strcmp(w,"user") == 0)
        {
            /*
            ** allow in only if the userid is one in the space separated list
            ** in the server cfg file. now t should have all the space 
            **  separated uids from cfg file 
            */
            /*
            ** go through all the uids and compare them with the uid
            ** from connection (which is in the LDAP server)
            */
            while (*t != '\0')
            {
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    m_rerr(r,cr->debug,"%s (%d) - ** Checking if  user \"%s\" is \"%s\"",MFL,user,w);

                    if (strncmp(w,user,strlen(w)) == 0)
                    {
                        m_rerr(r,cr->debug,"%s (%d) - OK user %s is authorized",MFL,user);
                        rcode=OK;
                        goto ExitProcessing;
                    }
                    else
                    {
                        m_rerr(r,cr->debug,"%s (%d) - ** user %s is is not in the list",MFL,user);
                    }
                } 
            }
            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }
        else if (strcmp(w,"filter") == 0)
        {
            int
                rc;
            char
                filter[256];

            /*
            ** We'll check if the user match the attributes in filter.
            ** We're going to use the user_dn obtained earlier, and we'll
            ** use the same open connection.
            */
            while (*t != '\0')
            {
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    ap_snprintf(filter,255,"%s",w);
                    m_rerr(r,cr->debug,"%s (%d) - Trying Filter: \"%s\"",MFL,filter);

                    /* Authorize */
                    rcode=mod_auth_ldap_mod_auth_ldap_check_user_authorization_real(r,
                            NULL,    /* group DN */
                            NULL,    /* Attr */
                            filter,  /* Filter */
                            filter); /* value */
                    if (rcode == OK)
                        goto ExitProcessing;
                    /* Authorize */
                }
            }   /* while */

            if (x >= reqs_arr->nelts -1)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            else
            {
                continue;
            }

            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }
        else if (strcmp(w,"group") == 0)
        {
            char
                *group_attrib=NULL,
                *group_name=NULL,
                szgdn[1024];

            group_attrib=ap_pstrdup(r->pool,w);
            rmallws(group_attrib);

            m_rerr(r,cr->debug,"%s (%d) - ldap attribute=%s",MFL,group_attrib);

            while (*t != '\0')
            {
                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes the string, so a group like "foo group" can not
                ** be used. -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    if (cr->group_attr)
                    {
                        m_rerr(r,cr->debug,
                            "%s (%d) - Group_Attr=%s",MFL,cr->group_attr);
                    }
                    ap_snprintf(szgdn,sizeof(szgdn)-1,"%s,%s",w,cr->base_dn);
                    m_rerr(r,cr->debug,"%s (%d) - Group DN:%s",MFL,szgdn);

                    /* Authorize */
                    rcode=mod_auth_ldap_mod_auth_ldap_check_user_authorization_real(r,
                            szgdn,  /* group DN */
                            NULL,   /* Attr */
                            NULL,   /* Filter */
                            w);     /* value */
                    if (rcode == OK)
                        goto ExitProcessing;
                    /* Authorize */
                }
            }   /* while */

            if (x >= reqs_arr->nelts -1)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            else
            {
                continue;
            }
            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }
        else
        {
            /*
            ** else allow in only if the ldap attribute is present in the LDAP
            ** server for the user
            */
            if (*w != '\0')
                attr=ap_pstrdup(r->pool,w);
            m_rerr(r,cr->debug,"%s (%d) - ldap attribute=%s",MFL,w);

            /* go through the attribute values and check if the user has it */
            while (*t != '\0')
            {
                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes string, so a attribute like "Foo org" can not
                ** be used.  -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    /* Authorize */
                    rcode=mod_auth_ldap_mod_auth_ldap_check_user_authorization_real(r,
                            NULL,  /* group DN */
                            attr,   /* Attr */
                            NULL,   /* Filter */
                            w);     /* value */
                    if (rcode == OK)
                        goto ExitProcessing;
                    /* Authorize */
                }
            }

            if (x >= reqs_arr->nelts -1)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto ExitProcessing;
            }
            else
            {
                continue;
            }

            rcode=HTTP_UNAUTHORIZED;
            goto ExitProcessing;
        }

        m_rerr(r,cr->debug,"%s (%d) - MMM x=%d,nelts-1=%d",MFL,x,(reqs_arr->nelts -1));
    }

ExitProcessing:

    if (user_dn != NULL)
    {
        m_rerr(r,cr->debug,"%s (%d) - ** Freeing dn: %s",user_dn);
        mod_auth_ldap_free_dn(user_dn);
    }

    /*
     * XXX crashes windows under heavy load in ssl. it crashes in 
     * ldap_simple_bind_s() routine. why?
     * so for now I'll commentit out for windows.. might cause handle leak!
     * muquit@muquit.com Dec-31-2002 
     */
    if (cr->ld != (LDAP *) NULL)
    {
#if defined(WIN32)
        if (cr->start_ssl)
            goto there;
#endif /* WIN32 */

    }
there:

    if (rcode != OK)
    {
        if (user)
            m_rerr(r,cr->debug,"%s (%d) - user \"%s\" is authenticated but not authorized. Hence not allowed!",MFL,user);

        /* delete the user from authen cache */
        mod_auth_ldap_delete_user_from_cache(r);
    }


    TMPM_UNLOCK;

    if (rcode == OK)
    {
        if (!cr->env_vars_set)
        {
            /* Set LDAP_USER environment variables in all cases */
            ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
            ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",MOD_AUTH_LDAP_VERSION_S);

            if (cr->set_pass_to_env && sent_pw)
            {
                ap_table_setn(r->subprocess_env,"LDAP_PASSWORD",sent_pw);
            }

            if (cr->uid_attr_alt)
            {
                mod_auth_ldap_set_ldap_user_alt(r);
            }

        }
    }
    if (rcode == HTTP_UNAUTHORIZED)
        ap_note_basic_auth_failure(r);

    return (rcode);
}



/* create per-server configuration */
static void *create_ldap_auth_server_config(apr_pool_t *p,server_rec *s)
{

    s_srv_conf=(LdapAuthSrvConf *) apr_pcalloc(p,sizeof(LdapAuthSrvConf));
    if (s_srv_conf == NULL)
        return(NULL);

    s_srv_conf->ld=NULL;

    s_srv_conf->use_cache=0;
    s_srv_conf->use_memory_cache=0;
    s_srv_conf->cache_ttl=300;
    s_srv_conf->use_disk_cache=0;

#ifdef USE_MEM_CACHE
    s_srv_conf->hash_table_size=0;
    s_srv_conf->hashp=NULL;
#endif /* USE_MEM_CACHE */

    s_srv_conf->ldap_persistent=1;
    s_srv_conf->support_nested_groups=0;


    return(s_srv_conf);
}

/* Show per-process stats. 
 * It's interesting only for a process. So when you request the stats
 * and if that process didn't go through the ldap auth process, the stats
 * will be meaningless.
 */

/*
<Location /mod_auth_ldap_stats>
   SetHandler mod_auth_ldap_stats
</Location>
*/
static int mod_auth_ldap_stats_handler(request_rec *r)
{
    char
        buf[BUFSIZ];

    CacheVal
        *cache_val=NULL;

    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *) ap_get_module_config(r->per_dir_config,&auth_ldap_module);


    if (ap_strcasecmp_match(r->handler,"mod_auth_ldap_stats") != 0)
    {
        return(DECLINED);
    }

    r->content_type="text/html";
    if (r->header_only)
    {
        return(OK);
    }

    m_rerr(r,1,"%s (%d) - Handler: %s",MFL,r->handler);

    TMPM_LOCK;

    ap_rputs(DOCTYPE_HTML_3_2
            "<html><head><title>mod_auth_ldap cache statistics</title></head>\n",r);

    ap_rputs("<body><h2>mod_auth_ldap v2.14 Cache Statistics</h2>\n",r);

    ap_rputs("<pre><font color=\"#000099\">\n",r);

#ifdef USE_MEM_CACHE
    if (s_srv_conf->use_memory_cache)
    {
        ap_snprintf(buf,sizeof(buf)-1,"(PID: %d) Hashtable size: %lu\n",getpid(),(unsigned long) ght_tablesize(s_srv_conf->hashp));
        ap_rputs(buf,r);
        ap_snprintf(buf,sizeof(buf)-1,"Hash count: %lu\n",(unsigned long) ght_size(s_srv_conf->hashp));
        ap_rputs(buf,r);
    }
    else
    {
        ap_snprintf(buf,sizeof(buf)-1,"cache stats available for memory cache only at this time\n");
        ap_rputs(buf,r);
    }
#endif
    ap_rputs("</pre>\n",r);

    ap_rputs("</body></html>\n",r);

    TMPM_UNLOCK;

    return(OK);
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_post_config(mod_auth_ldap_cache_post_config,
                         NULL,NULL,
                         APR_HOOK_MIDDLE);

    ap_hook_check_user_id(mod_auth_ldap_check_user_authentication,
                          NULL,NULL,
                          APR_HOOK_MIDDLE);

    ap_hook_auth_checker(mod_auth_ldap_check_user_authorization,
                         NULL,NULL,
                         APR_HOOK_MIDDLE);

    ap_hook_handler(mod_auth_ldap_stats_handler,
                        NULL,NULL,
                        APR_HOOK_MIDDLE);

}

/*--------------------------------------------------------------------------*/
/* Finally, the list of callback routines and data structures that          */
/* provide the hooks into our module from the other parts of the server.    */
/*--------------------------------------------------------------------------*/
module AP_MODULE_DECLARE_DATA auth_ldap_module =
{
    STANDARD20_MODULE_STUFF,
    create_ldap_auth_dir_config,    /* per-directory config creator */
    NULL,                           /* dir config merger */
    create_ldap_auth_server_config, /* server config creator */
    NULL,  /* server config merger */
    ldap_auth_cmds,                 /* command table */
    register_hooks                  /* register hooks */
};

/*
 * Usually I write code in a 80x50 xterm (in Unix and Windows). 
 * But this module is written in a window of size 134x68, because
 * logging routine has long message etc and looks ungly in small window. 
 *
 * If you modify the code, please follow my coding style:
 *
 *   -- No Tabs please, use four spaces instead. 
 *   -- ANSI style bracing like I do.
 *   -- Declare variables like I do. 
 *
 * - muquit@muquit.com Jan-07-2004 
 */

