#ifndef MOD_AUTH_LDAP_H
#define MOD_AUTH_LDAP_H

#include "apr_hooks.h"
#include "apr.h"
#include "apr_compat.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "apr_buckets.h"
#include "apr_md5.h"
#include "apr_network_io.h"
#include "apr_pools.h"
#include "apr_strings.h"
#include "apr_uri.h"
#include "apr_date.h"
#include "apr_fnmatch.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"

#include <stdio.h>

#include <lber.h>
#include <ldap.h>

#ifdef HAVE_SSL
#include <ldap_ssl.h>
#endif /* HAVE_SSL */

#include "sll.h"

#define MOD_AUTH_LDAP_VERSION_S "2.10"
#define MOD_AUTH_LDAP_SITE_URL  "http://muquit.com/"

/* non ssl port */
#ifndef LDAP_PORT
#define LDAP_PORT   389
#endif /* LDAP_PORT */

#define MML    __LINE__

typedef struct _conn_data
{
    LDAP
        *ld; /* the ldap connection */

#if APR_HAS_THREADS
    apr_thread_mutex_t
        *mutex; /* Used to serialize access to ld*/
#endif

    char
        *ldap_server;

    int
        ldap_port;
}conn_data;

typedef struct _ldap_auth_config_rec
{
    LDAP
        *ld; /* The ldap connection. Must be protected by a semaphore*/

#if APR_HAS_THREADS
    apr_thread_mutex_t
        *mutex; /* Used to serialize access to ld*/
#endif

    char
        *ldap_server,
        *base_dn,
        *uid_attr,
        *uid_attr_alt,
        *group_attr;

    char
        filter[512];

    char
        *user_dn;

    char
        *bind_dn,
        *bind_pass;

    int
        ldap_port,
        ldap_protocol_version, /* The ldap protocoll version to use*/
        ldap_connect_timeout;

    int
        env_vars_set,
        debug,
        auth_ldapauthoritative,
        support_nested_groups;

#ifdef HAVE_LDAP_START_TLS_S
    int
        auth_ldapstarttls;
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
    char
        *ca_cert_db_path; /* path of cert7.db in format used by netscape communicator*/
#endif /* HAVE_SSL */
    int
        start_ssl;


    int
        ldap_opt_deref;

    char
        *dummy;

    Sll
        *conn_ll;
} ldap_auth_config_rec;



#endif /* MOD_AUTH_LDAP_H */
